
open Cil
open FreeSafeTyUtil
open Format

let omaxid = ref 0
let get_omaxid () = !omaxid

type ovarid = (owner, int) tvarinfo

and owner =
  | OFrac of float
  | OVar of ovarid
  | OPlus of owner list
  | OScalar of float * owner
module OVS = Set.Make(
  struct
    type t = ovarid
    let compare v1 v2 = compare v1.id v2.id
  end
)


let rec find_owner o =
  let find ov =
    find (function OVar ov -> Some ov | _ -> None) (fun ov -> OVar ov) ov
  in
    match o with OVar ov -> fst (find ov) | _ -> o
and occur_check_owner ov o =
  let o = find_owner o in
  match o with OVar ov' when ov.id = ov'.id -> true
    | OPlus os -> List.fold_left (fun b o' -> b || occur_check_owner ov o') false os
    | OScalar _ -> failwith "occur_check_owner: cannot occur"
    | OVar _ | OFrac _ -> false
and union_owner loc ov o =
  merge (LocSet.singleton loc) occur_check_owner () ov o

let rec normalize_owner o =
  let o = find_owner o in
  let o' =
    match o with
      | OPlus os ->
	  let rec helper tobe_os (f, os) =
	    match tobe_os with 
	      | [] -> 
		  begin match f, os with
		    | 0.0, [] -> OFrac 0.0
		    | 0.0, [o] -> o
		    | 0.0, _ -> OPlus os
		    | _, [] -> OFrac f
		    | _, _ -> OPlus((OFrac f)::os)
		  end
	      | hd_o::tl_os ->
		  begin 
		    let hd_o = find_owner hd_o in
		      match hd_o with 
		      | OFrac f' -> helper tl_os (f +. f', os)
		      | OPlus os' -> helper (os' @ tl_os) (f, os)
		      | _ -> helper tl_os (f, hd_o::os) 
		  end
	  in
	    helper os (0.0, [])
      | OScalar(1.0, o) -> normalize_owner o
      | OScalar(0.0, o) -> OFrac 0.0
      | OScalar(f, OFrac f') -> OFrac (f *. f')
      | OFrac _ | OVar _ | OScalar _ -> o
  in
    (* printf "@[%a@] is normalized to @[%a@]@\n" print_owner o print_owner o'; *)
    o'

let rec compare_owner o1 o2 =
  let o1, o2 = normalize_owner o1, normalize_owner o2 in
    match o1, o2 with
      | OFrac f1, OFrac f2 -> compare f1 f2
      | OVar ov1, OVar ov2 -> compare ov1.id ov2.id
      | OPlus os1, OPlus os2 ->
	  compare_list compare_owner (List.sort compare_owner) os1 os2
      | OScalar(f1, o1), OScalar(f2, o2) ->
	  let cmp = compare f1 f2 in
	    if cmp = 0 then compare_owner o1 o2
	    else cmp
      | (OFrac _ | OVar _ | OPlus _ | OScalar _), 
	  (OFrac _ | OVar _ | OPlus _ | OScalar _) -> compare o1 o2

(* In printing owner by print_owner function,
   if print_only is Some ovs, then only ownership variables 
   contained in ovs is printed. *)
let print_only = ref None
let set_print_only ovs = print_only := Some ovs

(*
  struct
    type t = ovarid list
    let add ov os = if List.mem ov os then os else ov::os
    let union os1 os2 = List.fold_left (fun ovs ov -> add ov ovs) os1 os2
    let empty = []
    let singleton ov = [ov]
    let elements ovs = ovs
  end
*)
(* let glob_ovs = ref OVS.empty *)
let glob_ovs : (int, ovarid) Hashtbl.t = Hashtbl.create 40000
(* let get_glob_ovs () = !glob_ovs *)
let glob_ovs_iter f = Hashtbl.iter (fun _ ov -> f ov) glob_ovs
let get_ovi_from_id n = Hashtbl.find glob_ovs n
let new_ovar () = incr omaxid; let ovi = new_tvar !omaxid in Hashtbl.add glob_ovs ovi.id ovi; OVar ovi
let rec print_ovarid fmt ov = 
  match !print_only with
    | None ->
	print_tvar print_owner (fun fmt n -> fprintf fmt "%d" n) "o" fmt ov
    | Some ovs when OVS.mem ov ovs ->
	print_tvar print_owner (fun fmt n -> fprintf fmt "%d" n) "o" fmt ov
    | _ -> fprintf fmt ".."
and print_ovs fmt ovs =
  fprintf fmt "{@[%a@]}" (print_list print_ovarid ~sep:";") (OVS.elements ovs)
and  print_owner fmt o =
  match o with 
    | OFrac f -> fprintf fmt "%g" f
    | OVar v -> fprintf fmt "%a" print_ovarid v
    | OPlus ol -> fprintf fmt "%a" (print_list print_owner ~sep:"+") ol
    | OScalar(f, o) -> fprintf fmt "%g * %a" f print_owner o

let owner_zero = OFrac 0.0
let owner_one = OFrac 1.0
  
(*
let rec unify_owner o1 o2 =
  match normalize_owner (find_owner o1), normalize_owner (find_owner o2) with
    | OFrac f1, OFrac f2 when f1 = f2 -> ()
    | OVar ov1, OVar ov2 when ov1 == ov2 -> ()
    | OVar ov, ((OVar _ | OFrac _) as o) | ((OFrac _) as o), OVar ov ->
	ignore(union_owner ov o)
    | OPlus os, OFrac 0.0 | OFrac 0.0, OPlus os ->
	List.iter (fun o -> unify_owner o (OFrac 0.0)) os
    | OPlus _, OPlus _ -> ()
    | _ -> failwith "unify_owner: should not be passed"
*)

(* [o/ov]o' *)
let rec subst_ovar_owner ov o o' =
  let o_ret = 
    match o' with
      | OPlus os -> OPlus (List.rev_map (subst_ovar_owner ov o) os)
      | OVar ov' ->
	  if ov' = ov then o
	  else o'
      | OFrac _ -> o'
      | OScalar(f, o') -> OScalar(f, subst_ovar_owner ov o o')
  in
    normalize_owner o_ret
let subst_ovar_owner ov o o' =
  let o, o' = find_owner o, find_owner o' in
    subst_ovar_owner ov o o'

let rec rename_owner o =
  match o with
    | OFrac f -> OFrac f
    | OVar _ -> new_ovar ()
    | OPlus ol -> OPlus (List.rev_map rename_owner ol)
    | OScalar(f, o) -> OScalar(f, rename_owner o)
(* [XXX] should be merged with rename_owner *)
let rec rename_owner_renamefrac o =
  match o with
    | OFrac f -> new_ovar ()
    | OVar _ -> new_ovar ()
    | OPlus ol -> OPlus (List.rev_map rename_owner_renamefrac ol)
    | OScalar(f, o) -> OScalar(f, rename_owner_renamefrac o)
let owner_plus o1 o2 =
  match o1, o2 with
    | OFrac 0.0, o | o, OFrac 0.0 -> o
    | OFrac f1, OFrac f2 -> OFrac (f1 +. f2)
    | OVar _, OFrac _ | OFrac _, OVar _ | OVar _, OVar _ ->
	OPlus [o1; o2]
    | OPlus ol1, OPlus ol2 -> OPlus (ol1 @ ol2)
    | OPlus ol, o | o, OPlus ol -> OPlus (o::ol)
    | OScalar _, _ | _, OScalar _ -> failwith "owner_plus: cannot occur"

let tmaxid = ref (-1)
let get_tmaxid () = !tmaxid

type typ =
  | TSomething
  | TFun of typ list * tenv * typ list * tenv
      (* | TRef of typ * owner (* free *) * owner (* write *) *)
  | TRef of typ * owner (* free *) * (owner option) (* write *)
  (* [0107] | TRecord of (string * typ) list *)
  | TRecord of (string * typ * (owner option)) list
  | TVar of tvarid
  (* | TPlus of tvarid list *)
  | TTypPlus of typ list
  | TVarargs
and tvarid = (typ, int * Cil.typ) tvarinfo
and tenv = (varinfo * typ) list
(* let new_tvar () = incr tmaxid; TVar (new_tvar !tmaxid) *)

let find_typ typ =
  let find_typ tv =
    find (function TVar tv -> Some tv | _ -> None) (fun tv -> TVar tv) tv
  in
    match typ with TVar tv -> fst (find_typ tv) | _ -> typ
let rec occur_check_typ tv typ =
  let typ = find_typ typ in
    match typ with
    | TVar tv' when (fst tv.id) = (fst tv'.id) -> true
    | TFun(ts1, te1, ts2, te2) ->
	(List.fold_left
	   (fun b t -> b || occur_check_typ tv t)
	   false (ts1 @ ts2))
	||
	  (List.fold_left
	     (fun b (_, t) -> b || occur_check_typ tv t)
	     false (te1 @ te2))
    | TRef(t, _, _) -> occur_check_typ tv t
    | TRecord fts ->
	List.fold_left (fun b (_, t, _) -> b || occur_check_typ tv t) false fts
    | TTypPlus ts ->
	List.fold_left (fun b t -> b || occur_check_typ tv t) false ts
    | TVarargs | TVar _ | TSomething -> false
let union_typ loc tv typ =
  merge (LocSet.singleton loc) occur_check_typ () tv typ

let sort_tenv te = List.sort (fun (v1, _) (v2, _) -> compare v1.vname v2.vname) te
(*
let rec compare_typ typ1 typ2 =
  let typ1, typ2 = find_typ typ1, find_typ typ2 in
    match typ1, typ2 with
      | TSomething, TSomething -> 0
      | TRef(typ1, f1, w1), TRef(typ2, f2, w2) ->
	  compare_delayed_list
	    [(fun () -> compare_typ typ1 typ2);
	     (fun () -> compare_owner f1 f2);
	     (fun () -> compare_owner w1 w2)]
      | TFun(typs1pre, glob1pre, typs1post, glob1post),
	  TFun(typs2pre, glob2pre, typs2post, glob2post) ->
	  compare_delayed_list
	    [(fun () -> compare_list compare_typ (List.sort compare_typ) typs1pre typs2pre);
	     (fun () -> compare_list compare_typ (List.sort compare_typ) typs1post typs2post);
	     (fun () -> compare_tenv glob1pre glob2pre);
	     (fun () -> compare_tenv glob1post glob2post)]
      | TVar tv1, TVar tv2 -> compare (fst tv1.id) (fst tv2.id)
and compare_tenv te1 te2 =
  let compare_entry (x1, typ1) (x2, typ2) =
    compare_delayed_list
      [(fun () -> compare x1.vname x2.vname);
       (fun () -> compare_typ typ1 typ2)]
  in
    compare_list compare_entry sort_tenv te1 te2
*)
let fold_left_tenv f z (te : tenv) = List.fold_left f z te
let new_tvar ciltyp = incr tmaxid; TVar (new_tvar (!tmaxid, ciltyp))
let sort_fts fts = List.sort (fun (f1, _, _) (f2, _, _) -> compare f1 f2) fts
let rec print_tvarid fmt tv = 
  print_tvar print_typ (fun fmt (n, ciltyp) -> fprintf fmt "%d" n (*print_ciltyp ciltyp*)) "t" fmt tv
and print_typ (fmt : formatter) (typ : typ) : unit =
  match typ with
    | TSomething -> fprintf fmt "SOMETHING"
    | TFun(typs1, te1, typs2, te2) ->
	fprintf fmt "(@[%a@]|@\n[@[%a@]])@\n->@\n(@[%a@]|@\n[@[%a@]])"
	  (print_list print_typ ~sep:",@\n") typs1
	  print_tenv te1
	  (print_list print_typ ~sep:",@\n") typs2
	  print_tenv te2
    | TRef(typ', f, None) -> fprintf fmt "@[%a@] ref(@[%a@], _)" print_typ typ' print_owner f
    | TRef(typ', f, Some w) -> fprintf fmt "@[%a@] ref(@[%a@], @[%a@])" print_typ typ' print_owner f print_owner w
    | TRecord fts -> fprintf fmt "{@[%a@]}" print_fts (sort_fts fts)
    | TVar tvid ->
	fprintf fmt "@[%a@]" print_tvarid tvid
(*
    | TPlus tvids ->
	fprintf fmt "@[%a@]" (print_list print_tvarid ~sep:"+") tvids
*)
    | TTypPlus typs ->
	fprintf fmt "@{%a@]" (print_list print_typ ~sep:"+") typs
    | TVarargs -> ()
and print_fts fmt fts =
  fprintf fmt "@[%a@]" 
    (print_list (fun fmt (fn, t, o) -> 
		   match o with
		       None -> fprintf fmt "@[@[%s@](_):@[%a@]@])" fn print_typ t
		     | Some o -> fprintf fmt "@[@[%s@](@[%a@]):@[%a@]@])" fn print_owner o print_typ t) ~sep:";@\n") fts
and print_tenv fmt te =
  fprintf fmt "@[%a@]"
    (print_list (fun fmt (vi, typ) -> fprintf fmt "@[%a@]:@[%a@]" print_varinfo vi print_typ typ) ~sep:"@\n")
    te


let get_field fn fts =
  let _, typ, o = List.find (fun (fn', _, _) -> fn = fn') fts in
    typ, o


let rec rename_tenv ?(rename_fun=false) te = List.rev_map (fun (vi, typ) -> (vi, rename_typ ~rename_fun typ)) te
and rename_typ ?(rename_owner_fun=rename_owner) ?(rename_fun=false) typ =
  match typ with
    | TFun(pretyps, preglobs, posttyps, postglobs) ->
	if rename_fun then
	  TFun(List.map (rename_typ ~rename_owner_fun ~rename_fun) pretyps,
	       rename_tenv preglobs,
	       List.map (rename_typ ~rename_owner_fun ~rename_fun) posttyps,
	       rename_tenv postglobs)
	else
	  typ
    | TRef(typ', f, None) -> 
	TRef(rename_typ ~rename_owner_fun ~rename_fun typ', rename_owner_fun f, None)
    | TRef(typ', f, Some w) -> 
	TRef(rename_typ ~rename_owner_fun ~rename_fun typ', rename_owner_fun f, Some (rename_owner_fun w))
    | TRecord fts -> 
	TRecord (List.rev_map (function (fn, t, Some o) -> fn, rename_typ ~rename_owner_fun ~rename_fun t, Some (rename_owner_fun o)
				 | (fn, t, None) -> fn, rename_typ ~rename_owner_fun ~rename_fun t, None) fts)
    | TTypPlus typs -> TTypPlus (List.rev_map (rename_typ ~rename_owner_fun ~rename_fun) typs)
    | TVar tv -> new_tvar (snd tv.id)
    | TVarargs | TSomething (* | TPlus _ *) -> typ
let empty_tenv = []
let mem_tenv vi (te : tenv) =
  List.exists (fun (vi', _) -> vi.vname = vi'.vname) te
let add_tenv (vi, typ) te =
  if List.exists (fun (vi', _) -> vi.vname = vi'.vname) te then
    failwith "add_tenv: duplication in add_tenv"
  else
    (vi, typ)::te
let find_tenv vi te = List.assoc vi te
let disjoint_union_tenv te1 te2 = List.fold_left (fun te (vi, typ) -> add_tenv (vi, typ) te) te1 te2
let combine_tenv vars typs = List.combine vars typs
let rec tensor_typ loc typ1 typ2 =
  let typ1, typ2 = find_typ typ1, find_typ typ2 in
    match typ1, typ2 with
    | TSomething, TSomething -> TSomething

    (* Some-none mismatch CAN occur. Consider taking tensor of
       't[4034] ref('o[10181], 'o[10180]) and SOMETHING ref(1, _).
       This may occur when void* is involved. *)
    | TRef(typ1, f1, _), TRef(typ2, f2, None)
    | TRef(typ1, f1, None), TRef(typ2, f2, _) -> TRef(tensor_typ loc typ1 typ2, owner_plus f1 f2, None)
    | TRef(typ1, f1, Some w1), TRef(typ2, f2, Some w2) -> TRef(tensor_typ loc typ1 typ2, owner_plus f1 f2, Some (owner_plus w1 w2))
    | TRecord fts1, TRecord fts2 -> TRecord (tensor_fts loc fts1 fts2)
	(* | TVar tv1, TVar tv2 -> TTypPlus [typ1; typ2] *)
(* | TPlus tvs1, TPlus tvs2 -> TPlus (tvs1 @ tvs2) *)
(*
    | TPlus _, typ | typ, TPlus _ ->
	fstwarn "type variables in tensor_typ loc:@\n<typ1>@[%a@]@\n<typ2>@[%a@]@\n" print_typ typ1 print_typ typ2;
	typ
*)
    | TTypPlus typs, typ | typ, TTypPlus typs -> TTypPlus (typ::typs)
    | TVar _, _ | _, TVar _ -> TTypPlus [typ1; typ2]
    | TFun _, TFun _ ->
	(* if !FreeSafeTyFlag.debug then fstwarn "<tensor_typ loc>TFun: assumed to be equal.@\n"; *)
	typ1
    | _ -> 
	fstwarn "<typ1>@[%a@]@\n<typ2>@[%a@]@\n<loc>@[%a@]@\n" print_typ typ1 print_typ typ2 print_loc loc;
	failwith "tensor_typ loc: Not implemented yet"
and tensor_fts loc fts1 fts2 =
  let comparator (fn1, _, _) (fn2, _, _) = compare fn1 fn2 in
  let fts1, fts2 = List.sort comparator fts1, List.sort comparator fts2 in
    List.rev_map2 
      (fun (fn1, typ1, o1) (fn2, typ2, o2) ->
	 if fn1 = fn2 then 
	   match o1, o2 with
	     | Some o1', Some o2' -> (fn1, tensor_typ loc typ1 typ2, Some (owner_plus o1' o2'))
	     | None, None -> (fn1, tensor_typ loc typ1 typ2, None)
	     | _ ->
		 begin 
		   fstwarn "<fn1>@[%s@]@\n<fn2>@[%s@]@\n<typ1>@[%a@]@\n<typ2>@[%a@]@\n"
		     fn1 fn2 print_typ typ1 print_typ typ2;
		   failwith "tensor_fts loc: none-some does not match"
		 end
	 else
	   begin 
	     fstwarn "<fn1>@[%s@]@\n<fn2>@[%s@]@\n<typ1>@[%a@]@\n<typ2>@[%a@]@\n<loc>@[%a@]@\n"
	       fn1 fn2 print_typ typ1 print_typ typ2 print_loc loc;
	     failwith "tensor_fts loc: field names do not match"
	   end)
      fts1 fts2
and tensor_tenv loc te1 te2 =
  let te1, te2 = sort_tenv te1, sort_tenv te2 in
  let rec helper te1 te2 =
    match te1, te2 with
      | [], te | te, [] -> te
      | (v1, typ1)::tl1, (v2, typ2)::tl2 ->
	  let cmp = compare v1.vname v2.vname in
	    if cmp = 0 then (v1, tensor_typ loc typ1 typ2)::(helper tl1 tl2)
	    else if cmp < 0 then (v1, typ1)::(helper tl1 te2)
	    else (v2, typ2)::(helper te1 tl2)
  in
    helper te1 te2
let rec normalize_typ loc typ =
  let typ = find_typ typ in
    match typ with
    | ( TVarargs | TVar _ | TSomething) as typ' -> typ'
    | TFun(tspre, tepre, tspost, tepost) ->
	TFun(List.map (normalize_typ loc) tspre,
	     List.rev_map (fun (x, typ) -> x, normalize_typ loc typ) tepre,
	     List.map (normalize_typ loc) tspost,
	     List.rev_map (fun (x, typ) -> x, normalize_typ loc typ) tepost)
    | TRef(typ, f, None) -> TRef(normalize_typ loc typ, normalize_owner f, None)
    | TRef(typ, f, Some w) -> TRef(normalize_typ loc typ, normalize_owner f, Some (normalize_owner w))
    | TRecord fts -> 
	TRecord 
	  (List.rev_map (function (fn, typ, Some o) -> (fn, normalize_typ loc typ, Some (normalize_owner o))
			   | (fn, typ, None) -> (fn, normalize_typ loc typ, None)) fts)
    | TTypPlus [typ] -> normalize_typ loc typ
    | TTypPlus typs ->
	let typopt, tvs = 
	  List.fold_left 
	    (fun (typopt, tvs) typ' -> 
	       match typopt, normalize_typ loc typ' with
		 | _, TVar tv -> (typopt, tv::tvs)
		 | Some typ, newtyp -> (Some (tensor_typ loc typ newtyp), tvs)
		 | None, newtyp -> (Some newtyp, tvs))
	    (None, []) typs
	in
	  begin match typopt with
	      None -> TTypPlus (List.rev_map (fun tv -> TVar tv) tvs)
	    | Some typ -> TTypPlus(typ::(List.rev_map (fun tv -> TVar tv) tvs))
	  end
and normalize_tenv loc te = List.rev_map (fun (x, typ) -> (x, normalize_typ loc typ)) te

let rec destruct_reftyp loc typ =
  match typ with
    | TRef(typ', o, w) -> typ', o, w
    | TVar tv ->
	let ciltyp' = 
	  if isVoidPtrType (snd tv.id) then
	    voidPtrType
	  else
	    match unrollType (snd tv.id) with TPtr(t', _) | TArray(t', _, _) -> t'
	      | _ -> 
		  fstwarn "@[%a@]@\n" print_ciltyp (unrollType (snd tv.id));
		  failwith "destruct_reftyp: Not a ptr or an array type."
	in
	let t', f', w' = 
	  new_tvar ciltyp', 
	  new_ovar (), 
	  if isIntegralType ciltyp' || isArithmeticType ciltyp' || isFunctionType ciltyp' then
	    None
	  else
	    Some (new_ovar ()) 
	in
	let new_reftyp = TRef(t', f', w') in
	let _, b = union_typ loc tv new_reftyp in
	  assert(b);
	  t', f', w'
    (*| TVar _  *)| TFun _ | TSomething | TRecord _ | TTypPlus _ | TVarargs ->
	fstwarn "@[%a@]@\n" print_typ typ;
	failwith "destruct_reftyp: applied to non-ref type."
let get_glob_funtyp typ =
  match typ with TFun(_, preglob, _, postglob) -> preglob, postglob
    | _ ->
	fstwarn "get_glob_funtyp: [%a@]@\n" print_typ typ;
	failwith "get_glob_funtyp is applied to non-function type."
let rec typ_of_ciltyp ?(unfold=false) ciltyp =
  (* let expanded = ref [] in *)
  let rec helper unfold expanded ciltyp =
    match ciltyp with
	(* [XXX] *)
      | ciltyp when isVoidPtrType ciltyp -> new_tvar ciltyp
      | TEnum _ | TInt(_, _) | TFloat _ -> TSomething
      (* | TVoid _ -> new_tvar ()*)
      | Cil.TFun(TVoid _, None, _, _) ->
	  TFun([], [], [], [])
      | Cil.TFun(TVoid _, Some args, _, _) ->
	  let pretyps = List.map (fun (_, ciltyp, _) -> helper false expanded ciltyp) args in
	  let posttyps = List.map (fun (_, ciltyp, _) -> helper false expanded ciltyp) args in
	    TFun(pretyps, [], posttyps, [])
      | TArray(ciltyp', _, _) | TPtr(ciltyp', _) -> 
	  if isIntegralType ciltyp' || isArithmeticType ciltyp' || isFunctionType ciltyp' then
	    TRef(helper false expanded ciltyp', new_ovar (), None)
	  else
	    TRef(helper false expanded ciltyp', new_ovar (), Some (new_ovar ()))
      | TComp(ci, _) (*when ci.cstruct*) (* [XXX] Think about union *) ->
	  (* if List.mem ci.cname expanded then *)
	  if unfold && not (List.mem ci.cname expanded) then
	    TRecord 
	      (List.rev_map 
		 (fun fi -> 
		    fi.fname, 
		    helper false (ci.cname::expanded) fi.ftype, 
		    if isIntegralType fi.ftype || isArithmeticType fi.ftype || isFunctionType fi.ftype then None else Some (new_ovar ()))
		 ci.cfields)
	  else
	    new_tvar ciltyp
      | TNamed(ti, _) ->
	  (* if List.mem ti.tname expanded then *)
	  if unfold && not (List.mem ti.tname expanded) then
	    helper false (ti.tname::expanded) ti.ttype
	  else
	    new_tvar ciltyp
(*
		else
*)
      | TBuiltin_va_list _ -> TVarargs
      | _ -> 
	  fstwarn "@[%a@]@\n" print_ciltyp ciltyp;
	  failwith "typ_of_ciltyp, helper: not implemented"
  in
    helper unfold [] ciltyp

let globalTEnv = ref empty_tenv
let find_global_env v = find_tenv v !globalTEnv

class setupEnvVisitor : cilVisitor = object(self)
  inherit nopCilVisitor
   
  method private processDec vi =
    (* fstwarn "processDec: adding %s@\n%!" vi.vname; *)
    if mem_tenv vi !globalTEnv then
      ()
    else
      begin 
	let typ = typ_of_ciltyp vi.vtype in
	  globalTEnv := add_tenv (vi, typ) !globalTEnv;
	  (* fstwarn "@[%s@] : @[%a@]@ is added to the global type env.@\n" vi.vname print_typ typ *)
      end

  method vvdec vi = self#processDec vi; DoChildren

  method vfunc fd = self#processDec fd.svar; SkipChildren
end

let rec setupEnv f =
  let setupEnvVisitor = new setupEnvVisitor in
    visitCilFile setupEnvVisitor f;
    let globvars = 
      List.filter 
	(fun (vi, typ) ->
	   match typ with
	     | TVarargs | TSomething | TFun _ | TVar _ (* | TPlus _ *) -> false
	     | TRef _ | TRecord _ -> true
	     | TTypPlus _ -> failwith "setupEnv: Cannot occur")
	!globalTEnv
    in
    let rec add_globvars_to_funtyp globvars typ =
      match typ with
	| TRef(typ', f, w) -> TRef(add_globvars_to_funtyp globvars typ', f, w)
	| TRecord fts -> TRecord (List.rev_map (fun (fn, typ, o) -> (fn, add_globvars_to_funtyp globvars typ, o)) fts)
	| TFun(pretyps, [], posttyps, []) ->
	    TFun(List.map (add_globvars_to_funtyp globvars) pretyps,
		 rename_tenv ~rename_fun:true globvars,
		 List.map (add_globvars_to_funtyp globvars) posttyps,
		 rename_tenv ~rename_fun:true globvars)
	| TFun _ -> failwith "add_globvars_to_funtyp: do not apply me twice to a type"
	| TVarargs (* | TPlus _ *) | TVar _ | TSomething -> typ
	| TTypPlus _ -> failwith "add_globvars_to_funtyp: cannot occur"
    in
      globalTEnv := List.rev_map (fun (vi, typ) -> (vi, add_globvars_to_funtyp globvars typ)) !globalTEnv
type loc = Cil.location

(* If 'fundef option' is Some fd, then the constraint is generated in
   the global function fd. Otherwise, it is generated somewhere
   outside global function definitions. *)
type constr = 
    CLeOwner of owner * owner * loc
  | CEqOwner of owner * owner * loc
  | CSubtype of typ * typ * loc
  | CEmpty of tvarid * loc
  | CWfTyp of typ * loc

let rec compare_constr c1 c2 =
  match c1, c2 with
    | CEqOwner(o11, o12, _), CEqOwner(o21, o22, _)
    | CLeOwner(o11, o12, _), CLeOwner(o21, o22, _) -> 
	let cmp = compare_owner (find_owner o11) (find_owner o21) in
	  if cmp = 0 then compare_owner (find_owner o12) (find_owner o22)
	  else cmp
    | _ -> compare c1 c2

let rec get_freeovids_owner o =
  match find_owner o with
    | OVar ovid -> OVS.singleton ovid
    | OScalar(_, o) -> get_freeovids_owner (find_owner o)
	(* | OPlus(o1, o2) -> OVS.union (get_freeovids_owner o1) (get_freeovids_owner o2) *)
    | OPlus os -> List.fold_left (fun os o -> OVS.union os (get_freeovids_owner (find_owner o))) OVS.empty os
    | OFrac _ -> OVS.empty
and get_freeovids_typ typ =
  match typ with
    | TFun(ts1, te1, ts2, te2) ->
	let ovs1, ovs2 =
	  List.fold_left
	    (fun ovs t -> OVS.union ovs (get_freeovids_typ t))
	    OVS.empty
	    ts1,
	  List.fold_left
	    (fun ovs t -> OVS.union ovs (get_freeovids_typ t))
	    OVS.empty
	    ts2
	in
	let ovs3, ovs4 =
	  fold_left_tenv
	    (fun ovs (_, t) -> OVS.union ovs (get_freeovids_typ t))
	    OVS.empty
	    te1,
	  fold_left_tenv
	    (fun ovs (_, t) -> OVS.union ovs (get_freeovids_typ t))
	    OVS.empty
	    te2
	in
	  OVS.union (OVS.union (OVS.union ovs1 ovs2) ovs3) ovs4
    | TRef(typ, o1, Some o2) ->
	OVS.union (OVS.union (get_freeovids_typ typ) (get_freeovids_owner o1)) (get_freeovids_owner o2)
    | TRef(typ, o1, None) ->
	OVS.union (get_freeovids_typ typ) (get_freeovids_owner o1)
    | TRecord fts ->
	List.fold_left
	  (fun ovs (_, t, w) ->
	     match w with
		 Some w ->
		   OVS.union (OVS.union ovs (get_freeovids_typ t)) (get_freeovids_owner w)
	       | None -> OVS.union ovs (get_freeovids_typ t))
	  OVS.empty
	  fts
    | TTypPlus ts ->
	List.fold_left
	  (fun ovs t -> OVS.union ovs (get_freeovids_typ t))
	  OVS.empty ts
    | TVarargs | TSomething | TVar _ -> OVS.empty
and get_freeovids_tenv te =
  List.fold_left (fun ovs (_, typ) -> OVS.union (get_freeovids_typ typ) ovs) OVS.empty te

let rec print_constr fmt c =
  match c with
    (* | CLeOwner(OFrac f1, OFrac f2) when f1 <= f2 -> () *)
    | CLeOwner(o1, o2, loc) -> 
	begin
	  match !print_only with
	    | None -> fprintf fmt "@[%a@] <= @[%a@](@[%a@])" print_owner o1 print_owner o2 print_loc loc
	    | Some ovs when OVS.subset (OVS.union (get_freeovids_owner o1) (get_freeovids_owner o2)) ovs ->
		fprintf fmt "@[%a@] <= @[%a@](@[%a@])" print_owner o1 print_owner o2 print_loc loc
	    | _ -> ()
	end
    | CEqOwner(o1, o2, loc) -> fprintf fmt "@[%a@] = @[%a@](@[%a@])" print_owner o1 print_owner o2 print_loc loc
    | CSubtype(typ1, typ2, loc) -> fprintf fmt "@[%a@] <= @[%a@](@[%a@])"
	print_typ typ1 print_typ typ2 print_loc loc
    | CEmpty(tvid, loc) -> fprintf fmt "empty(@[%a@])(@[%a@])" print_tvarid tvid print_loc loc
    | CWfTyp(typ, loc) -> fprintf fmt "wf(@[%a@])(@[%a@])" print_typ typ print_loc loc
module CS = 
  Set.Make(
    struct
      type t = constr
      let compare c1 c2 = 
	(* fstwarn "<compare>@[%a@],@[%a@]@\n%!" print_constr c1 print_constr c2; *)
	compare_constr c1 c2
    end
  )


let rec get_freeovids_constr c =
  match c with
    | CLeOwner(o1, o2, _) (*| CLtOwner(o1, o2) *) -> 
	OVS.union (get_freeovids_owner (find_owner o1)) (get_freeovids_owner (find_owner o2))
    | CEqOwner(o1, o2, _) ->
	OVS.union (get_freeovids_owner (find_owner o1)) (get_freeovids_owner (find_owner o2))
    | CEmpty(tv, _) -> get_freeovids_typ (find_typ (TVar tv))
    | CSubtype(typ1, typ2, _) ->
	OVS.union (get_freeovids_typ typ1) (get_freeovids_typ typ2)
    | CWfTyp(typ, _) -> get_freeovids_typ (find_typ typ)
	  (* | _ -> failwith "get_freeovids_constr: not implemented" *)
and get_freeovids_cs cs =
  CS.fold (fun c s -> OVS.union (get_freeovids_constr c) s) cs OVS.empty

(*
  struct
    type t = constr list
    let empty = []
    let add c cs = if List.mem c cs then cs else c::cs
    let union cs1 cs2 = List.fold_left (fun cs c -> add c cs) cs1 cs2
    let singleton c = [c]
    let elements cs = cs
    let fold = List.fold_right
    let iter = List.iter
  end
*)


type constr_set = CS.t



let rec print_cs fmt cs =
  fprintf fmt "@[%a@]" (print_list print_constr ~sep:"@ @ @ @ ") (CS.elements cs)
let empty_cs = CS.empty
let union_cs cs1 cs2 =
  (* List.fold_left (fun cs c -> if List.mem c cs then cs else c::cs) cs1 cs2 *)
  (* fstwarn "<union>@[%a@],@\n@[%a@]@\n%!" print_cs cs1 print_cs cs2;*)
  let cs = CS.union cs1 cs2 in
    (* fstwarn "<union_end>@\n%!"; *) cs
let merge_cs csl = (* failwith "merge_cs: not implemented yet." *)
  (* fstwarn "<merge>@\n%!"; *)
  let cs = List.fold_left union_cs empty_cs csl in
    (* fstwarn "<merge_end>@\n%!"; *) cs
let epsilon_owner = OFrac (0.00048828125)
let le_owner loc o1 o2 = 
  let o1, o2 = find_owner o1, find_owner o2 in
  let _ =
    match o1, o2 with OFrac 1.0, OFrac 0.0 -> ()
      | _ -> ()
  in
  let cs =
    match o1, o2 with
	(* | OVar _, OFrac 1.0 -> CS.empty *)
      | OFrac f1, OFrac f2 when f1 <= f2 -> CS.empty
      | _, _ when o1 = o2 -> CS.empty
      | _ ->
	  CS.singleton (CLeOwner(o1, o2, loc))
  in
(*
    fstdebug "<le_owner>@[%a@] <= @[%a@]@\n" print_owner o1 print_owner o2;
    fstdebug "<cs>@[%a@]@\n" print_cs cs;
*)
    cs
let eq_owner loc o1 o2 = 
  let o1, o2 = find_owner o1, find_owner o2 in
  (* let _ = fstdebug "<eqowner>@[@[%a@]@\n@[%a@]@]@\n" print_owner o1 print_owner o2 in *)
  let cs = 
    match o1, o2 with
      | OFrac f1, OFrac f2 when f1 = f2 -> empty_cs
      | OVar ov, o | o, OVar ov ->
	  (* unify_owner o1 o2; *)
	  if !FreeSafeTyFlag.unification then
	    let _, b = union_owner loc ov o in
	      if b then empty_cs else merge_cs [le_owner loc o1 o2; le_owner loc o2 o1]
	  else
	    merge_cs [le_owner loc o1 o2; le_owner loc o2 o1]
      | _ ->
	  merge_cs [le_owner loc o1 o2; le_owner loc o2 o1]
  in
  (* let _ = fstdebug "<o1>@[%a@]@\n<o2>@[%a@]@\n<cs>@[%a@]@\n" print_owner o1 print_owner o2 print_cs cs in *)
    cs

		  
let lt_owner loc o1 o2 = CS.singleton (CLeOwner(epsilon_owner, o2, loc))
let rec sub_type loc typ1 typ2 : constr_set =
  let typ1, typ2 = find_typ typ1, find_typ typ2 in
  let cs =
  match typ1, typ2 with
    | TSomething, TSomething | TVarargs, TVarargs -> empty_cs
    | TFun(preargs1, preglobs1, postargs1, postglobs1), TFun(preargs2, preglobs2, postargs2, postglobs2) ->
	let c1 = List.rev_map2 (fun typ1 typ2 -> eq_typ loc typ1 typ2) preargs1 preargs2 in
	let c2 = List.rev_map2 (fun typ1 typ2 -> eq_typ loc typ1 typ2) postargs1 postargs2 in
	let c3 = union_cs (sub_tenv loc preglobs1 preglobs2) (sub_tenv loc preglobs2 preglobs1) in
	let c4 = union_cs (sub_tenv loc postglobs1 postglobs2) (sub_tenv loc postglobs2 postglobs1) in
	  merge_cs [merge_cs c1; merge_cs c2; c3; c4]
    | TRef(typ1, f1, Some w1), TRef(typ2, f2, Some w2) ->
	merge_cs [sub_type loc typ1 typ2; eq_owner loc f1 f2; (*eq_owner loc w2 w1*)le_owner loc w2 w1]

    (* For a remark of some-none mismatch, see the comment in tesor_typ *)
    | TRef(typ1, f1, _), TRef(typ2, f2, None) | TRef(typ1, f1, None), TRef(typ2, f2, _) ->
	merge_cs [sub_type loc typ1 typ2; eq_owner loc f1 f2; (*eq_owner loc w2 w1*)(*le_owner loc w2 w1*)]
    | TRecord fts1, TRecord fts2 -> sub_fts loc fts1 fts2
	(* | TPlus _, _ | _, TPlus _ -> [CSubtype(typ1, typ2)] *)
    | TVar tv1, TVar tv2 when tv1 == tv2 -> empty_cs
    | (TVar _ | TTypPlus _), _ | _, (TVar _ | TTypPlus _) -> CS.singleton (CSubtype(typ1, typ2, loc))
    | _ -> 
	fstwarn "<typ1>@[%a@]@\n<typ2>@[%a@]@\n" print_typ typ1 print_typ typ2;
	failwith "sub_type: type error"
  in
(*
    fstdebug "<sub_type>@[%a@] <= @[%a@]@\n" print_typ typ1 print_typ typ2;
    fstdebug "<cs>@[%a@]@\n" print_cs cs;
*)
    cs
and sub_fts loc fts1 fts2 : constr_set =
  let comparator (fn1, _, _) (fn2, _, _) = compare fn1 fn2 in
  let fts1, fts2 = List.sort comparator fts1, List.sort comparator fts2 in
  let c =
    List.fold_left2
      (fun cs (fn1, typ1, o1) (fn2, typ2, o2) ->
	 if fn1 = fn2 then
	   match o1, o2 with
	     | Some o1, Some o2 ->
		 merge_cs [cs; sub_type loc typ1 typ2; eq_owner loc o1 o2]
	     | None, None -> cs
	     | _ -> 
		 fstwarn "<fn1>@[%s@]@\n<fn2>@[%s@]@\n<typ1>@[%a@]@\n<typ2>@[%a@]@\n"
		   fn1 fn2 print_typ typ1 print_typ typ2;
		 failwith "sub_fts: some-none does not match."
	 else
	   (fstwarn "<fn1>@[%s@]@\n<fn2>@[%s@]@\n<typ1>@[%a@]@\n<typ2>@[%a@]@\n<loc>@[%a@]@\n"
	      fn1 fn2 print_typ typ1 print_typ typ2 print_loc loc;
	    failwith "sub_fts: field names do not match.@\n"))
      empty_cs fts1 fts2
  in
(*
    fstdebug "<sub_fts>@\n<fts1>@[%a@]@\n<fts2>@[%a@]@\n<c>@[%a@]@\n"
      print_fts fts1 print_fts fts2 print_cs c;
*)
    c

and eq_typ loc typ1 typ2 =
  let typ1, typ2 = find_typ typ1, find_typ typ2 in
    (* let _ = fstdebug "<eqtyp>@[<typ1>@[%a@]@\n<typ2>@[%a@]@]@\n" print_typ typ1 print_typ typ2 in *)
  (* let _ = fstdebug "<eqtyp>@\n<typ1>@[%a@]@\n<typ2>@[%a@]@\n" print_typ typ1 print_typ typ2 in *)
  let cs =
    match typ1, typ2 with
    | TSomething, TSomething -> empty_cs
    | TVar tv1, TVar tv2 when tv1 == tv2 -> empty_cs
    | ((TVar tv) as typ1), (typ as typ2) | (typ as typ1), ((TVar tv) as typ2) ->
	if !FreeSafeTyFlag.unification then
	  let _, b = union_typ loc tv typ in
	    if b then empty_cs else merge_cs [sub_type loc typ1 typ2; sub_type loc typ2 typ1]
	else
	  merge_cs [sub_type loc typ1 typ2; sub_type loc typ2 typ1]
    | TRef(t1, f1, Some w1), TRef(t2, f2, Some w2) ->
	merge_cs [eq_typ loc t1 t2; eq_owner loc f1 f2; eq_owner loc w1 w2]

    (* See impl. of tensor_typ for a comment on some-none mismatch *)
    | TRef(t1, f1, None), TRef(t2, f2, _) | TRef(t1, f1, _), TRef(t2, f2, None) ->
	merge_cs [eq_typ loc t1 t2; eq_owner loc f1 f2(*; eq_owner loc w1 w2*)]
    | TFun(tspre1, tepre1, tspost1, tepost1), TFun(tspre2, tepre2, tspost2, tepost2) ->
	let tepre1, tepre2, tepost1, tepost2 = sort_tenv tepre1, sort_tenv tepre1, sort_tenv tepost1, sort_tenv tepost2 in
	  merge_cs [safe_fold_left (fun cs t1 t2 -> merge_cs [cs; eq_typ loc t1 t2]) empty_cs tspre1 tspre2;
		    safe_fold_left (fun cs t1 t2 -> merge_cs [cs; eq_typ loc t1 t2]) empty_cs tspost1 tspost2;
		    safe_fold_left (fun cs (x1, t1) (x2, t2) -> assert(x1.vname = x2.vname); merge_cs [cs; eq_typ loc t1 t2]) 
		      empty_cs tepre1 tepre2;
		    safe_fold_left (fun cs (x1, t1) (x2, t2) -> assert(x1.vname = x2.vname); merge_cs [cs; eq_typ loc t1 t2]) 
		      empty_cs tepost1 tepost2]
    | TRecord fts1, TRecord fts2 ->
	let fts1, fts2 = sort_fts fts1, sort_fts fts2 in
	  List.fold_left2 
	    (fun cs (fn1, typ1, w1) (fn2, typ2, w2) -> 
	       match w1, w2 with
		   Some w1, Some w2 ->
		     merge_cs [cs; eq_typ loc typ1 typ2; eq_owner loc w1 w2]
		 | None, None ->
		     merge_cs [cs; eq_typ loc typ1 typ2]
		 | _ ->
		     failwith "eq_typ: some-none does not match")
	    empty_cs fts1 fts2
    | ((TTypPlus _) as typ1), typ2 | typ1, ((TTypPlus _) as typ2) -> merge_cs [sub_type loc typ1 typ2; sub_type loc typ2 typ1]
    | typ1, typ2 ->
	fstwarn "<typ1>@\n@[%a@]@\n<typ2>@[%a@]@\n<loc>@[%a@]@\n" print_typ typ1 print_typ typ2 print_loc loc;
	failwith "eq_typ: type error"
  in
  (* let _ = fstdebug "<eqtyp:after>@\n<typ1>@[%a@]@\n<typ2>@[%a@]@\n<cs>@[%a@]@\n" print_typ typ1 print_typ typ2 print_cs cs in *)
    cs

and sub_tenv loc te1 te2 : constr_set =
  let te1, te2 = sort_tenv te1, sort_tenv te2 in
  let rec helper cs te1 te2 =
    match te1, te2 with
      | [], [] -> cs
      | (v1, typ1)::tl1, (v2, typ2)::tl2 when v1.vname = v2.vname ->
	  helper (union_cs (sub_type loc typ1 typ2) cs) tl1 tl2
      | _ ->
	  fstwarn "te1 = @[%a@]@\nte2 = @[%a@]@\n<loc>@[%a@]@\n" print_tenv te1 print_tenv te2 print_loc loc;
	  failwith "sub_tenv: two type environments are incompatible."
  in
    helper empty_cs te1 te2
and eq_tenv loc te1 te2 : constr_set =
  let te1, te2 = sort_tenv te1, sort_tenv te2 in
  let rec helper cs te1 te2 =
    match te1, te2 with
      | [], [] -> cs
      | (v1, typ1)::tl1, (v2, typ2)::tl2 when v1.vname = v2.vname ->
	  helper (union_cs (eq_typ loc typ1 typ2) cs) tl1 tl2
      | _ ->
	  fstwarn "te1 = @[%a@]@\nte2 = @[%a@]@\n" print_tenv te1 print_tenv te2;
	  failwith "sub_tenv: two type environments are incompatible."
  in
    helper empty_cs te1 te2
  
let rec empty_typ loc typ =
  match typ with
    | TVarargs | TSomething | TFun _ (* | TVar _ *) -> empty_cs
    | TRef(typ', f, w) -> merge_cs [empty_typ loc typ'; eq_owner loc f owner_zero(*; eq_owner loc w owner_zero*)]
    | TRecord fts -> merge_cs (List.rev_map (fun (f, typ, _) -> empty_typ loc typ) fts)
    | TVar tvid -> CS.singleton (CEmpty(tvid, loc))
    | TTypPlus typs -> merge_cs (List.rev_map (empty_typ loc) typs)
let empty_cs_tenv loc te =
  List.fold_left (fun cs (_, typ) -> union_cs (empty_typ loc typ) cs) empty_cs te
let rec apply_typ loc funtyp preargs postargs =
(*
  fstdebug "Processing applytyp:@\n<funtyp>@[%a@]@\n<preargs>@[%a@]@\n<postargs>@[%a@]@\n"
    print_typ funtyp (print_list print_typ ~sep:"@\n") preargs (print_list print_typ ~sep:",@\n") postargs;
*)
  let cs =
    match funtyp with
      | TFun(preargs', _, postargs', _) ->
	  let c1 = (* List.fold_left2 *) safe_fold_left
	    (fun c typ' typ ->
	       match typ' with
		   TVar _ -> c
		 | _ -> merge_cs [c; eq_typ loc typ typ'])
	      empty_cs preargs' preargs 
	  in
	let c2 = (* List.fold_left2 *) safe_fold_left
	  (fun c typ' typ ->
	     match typ' with
		 TVar _ -> c
	       | _ -> merge_cs [c; eq_typ loc typ typ'])
	    empty_cs postargs' postargs 
	in
	  merge_cs [c1; c2]
      | _ ->
	  fstwarn "funtyp: @[%a@]@\npreargs: @[%a@]@\npostargs: @[%a@]@\n"
	    print_typ funtyp
	    (print_list print_typ ~sep:", ") preargs
	    (print_list print_typ ~sep:", ") postargs;
	  failwith "apply_typ is applied to non-function type."
  in
(*
    fstdebug "After applytyp:@\n<funtyp>@[%a@]@\n<preargs>@[%a@]@\n<postargs>@[%a@]@\n<cs>@[%a@]@\n"
      print_typ funtyp (print_list print_typ ~sep:"@\n") preargs (print_list print_typ ~sep:",@\n") postargs print_cs cs;
*)
    cs

(* let wf_cnt = ref 0 *)
let rec wf_te loc te = merge_cs (List.rev_map (fun (_, typ) -> wf_typ loc typ) te)
and wf_typ loc typ =
  (* let _ = printf "<WF_TYP:%d>@[%a@]@\n" !wf_cnt print_typ typ in *)
  (* let oldcnt = !wf_cnt in *)
  (* let _ = incr wf_cnt in *)
  let cs =
    let typ = normalize_typ loc (find_typ typ) in
      match typ with
	| TVarargs | TSomething -> empty_typ loc typ
	| TTypPlus _ | TVar _ -> CS.singleton (CWfTyp(typ, loc))
	| TFun(tspre, tepre, tspost, tepost) ->
	    let c1 = merge_cs (List.rev_map (wf_typ loc) tspre) in
	    let c2 = merge_cs (List.rev_map (wf_typ loc) tspost) in
	    let c3 = wf_te loc tepre in
	    let c4 = wf_te loc tepost in
	      merge_cs [c1; c2; c3; c4]
      | TRef(typ', f, Some w) ->
	  merge_cs [wf_typ loc typ'; le_owner loc f (OFrac 1.0); le_owner loc w (OFrac 1.0)]
      | TRef(typ', f, None) ->
	  merge_cs [wf_typ loc typ'; le_owner loc f (OFrac 1.0)]
      | TRecord fts -> 
	  merge_cs 
	    (List.rev_map (function (_, typ, Some o) -> merge_cs [wf_typ loc typ; le_owner loc o (OFrac 1.0)]
			     | (_, typ, None) ->  wf_typ loc typ) fts)
  in
  (* let _ = printf "<WF_TYP LOC:CS:%d>@[%a@]@\n" oldcnt print_cs cs in *)
    cs

let rec normalize_cs (cs : constr_set) : constr_set =
  CS.fold
    (fun c cs ->
       match c with
	 | CLeOwner(o1, o2, loc) ->
	     let o1, o2 = normalize_owner (find_owner o1), normalize_owner (find_owner o2) in
	       begin
		 match o1, o2 with
		   | OFrac f1, OFrac f2 when f1 <= f2 -> cs
		   | OVar ov1, OVar ov2 when ov1.id = ov2.id -> cs
		       
		   (* Don't remove ov <= 1.0. This is not trivial if
		      ov is substituted afterward. *)
		   | OFrac 0.0, OVar _ (*| OVar _, OFrac 1.0*) -> cs
		   | OFrac 0.0, OPlus _ -> cs
		   | _ -> CS.union (le_owner loc o1 o2) cs
	       end
	 | CEmpty(tv, loc) -> CS.union (empty_typ loc (find_typ (TVar tv))) cs
	 | CSubtype(typ1, typ2, loc) -> CS.union (sub_type loc (normalize_typ loc typ1) (normalize_typ loc typ2)) cs
	 | CWfTyp(typ, loc) -> CS.union (wf_typ loc (normalize_typ loc typ)) cs
	 | CEqOwner(o1, o2, loc) -> CS.add c cs
	 (*| _ -> CS.add c cs*))
    cs
    empty_cs
  (* cs *)

exception ExpansionSaturated

(* Expands subtyping constraint in which (addition of) tvars and
   non-tvar types are mixed.  Doesn't expand tvar whose id is more than
   or equal to maxid. *)
let rec expand_tvar_and_flatten loc maxid cs : constr_set * bool =
  let rec no_need_to_expand typ =
    match typ with TVar _ -> true
      | TTypPlus ts -> List.fold_left (fun b typ -> if b then no_need_to_expand typ else b) true ts
      | _ -> false
  in
  let cs_r = ref CS.empty in
  let dirty = ref false in
    CS.iter
      (fun c ->
	 match c with
	   | CLeOwner _ | CEqOwner _ -> cs_r := CS.add c !cs_r
	   | CSubtype(typ1, typ2, loc) ->
	       let typ1, typ2 = find_typ typ1, find_typ typ2 in
		 if no_need_to_expand typ1 && no_need_to_expand typ2 then
		   cs_r := CS.add (CSubtype(typ1, typ2, loc)) !cs_r
		 else
		   let (typ1, dirty1), (typ2, dirty2) = 
		     expand_tvar_and_flatten_typ loc maxid typ1, expand_tvar_and_flatten_typ loc maxid typ2 in
		   let template_typ = find_template [typ1; typ2] in
		   let (typ1, dirty3), (typ2, dirty4) = 
		     expand_tvar_and_flatten_typ ~template:template_typ loc maxid typ1, 
		     expand_tvar_and_flatten_typ ~template:template_typ loc maxid typ2 
		   in
		     dirty := !dirty || dirty1 || dirty2 || dirty3 || dirty4;
		     cs_r := CS.union (sub_type loc typ1 typ2) !cs_r
	   | CEmpty(tv, loc) ->
	       let typ, dirty' = expand_tvar_and_flatten_typ loc maxid (find_typ (TVar tv)) in
		 dirty := !dirty || dirty';
		 cs_r := CS.union (empty_typ loc typ) !cs_r		 
	   | CWfTyp(typ, loc) ->
	       let typ, dirty' = expand_tvar_and_flatten_typ loc maxid typ in
		 dirty := !dirty || dirty';
		 cs_r := CS.union (wf_typ loc typ) !cs_r)
      cs;
    !cs_r, !dirty
and expand_tvar_and_flatten_typ ?(template=None) loc maxid typ : typ * bool =
  let typ = normalize_typ loc (find_typ typ) in
    (* printf "<EXPAND>@[%a@]@\n" print_typ typ; *)
  let rettyp =
    match typ with
      | TSomething | TFun _ | TRef _ | TRecord _ | TVarargs -> typ, false
      | TTypPlus [] -> failwith "expand_tvar_and_flatten_typ: empty TTypPlus"
      | TTypPlus [typ] ->
	  expand_tvar_and_flatten_typ ~template:template loc maxid typ
      | TTypPlus (hd::tl) ->
	  let expand_with_template template =
	    List.fold_left
	      (fun (typ1, dirty1) typ2 ->
		 let expanded_typ2, dirty2 = expand_tvar_and_flatten_typ ~template:template loc maxid typ2 in
		   (tensor_typ loc typ1 expanded_typ2), (dirty1 || dirty2))
	      (expand_tvar_and_flatten_typ ~template:template loc maxid hd)
	      tl
	  in
	    begin match template with
	      | None ->
		  let template = find_template (hd::tl) in
		    expand_with_template template
	      | _ ->
		  expand_with_template template
	    end
      | TVar tv ->
	  begin match template with
	    | None when (isVoidPtrType (snd tv.id)) ->  
		(* Control cannot reach here because a check by
		   no_need_to_expand should have succeeded if this
		   function is called. (But assertion failed... why?)
		*)
		(* [XXX] *)
		(* assert(false);*)
		typ, false
	    | (Some templtyp) when (fst tv.id) < maxid && (isVoidPtrType (snd tv.id)) ->  
		let rettyp = rename_typ ~rename_owner_fun:rename_owner_renamefrac ~rename_fun:false templtyp in
		let _, b = union_typ loc tv rettyp in
		  if b then
		    rettyp, true
		  else
		    failwith "expand_with_template: expansion of recursive type."
	    | _ when (fst tv.id) < maxid ->
		let rettyp = typ_of_ciltyp ~unfold:true (snd tv.id) in
		  (* let _ = fstdebug "@[%a@] is expanded to @[%a@]@\n" print_typ typ print_typ rettyp in *)
		let _, b = union_typ loc tv rettyp in
		  if b then
		    rettyp, true
		  else
		    failwith "expand_with_template: expansion of recursive type."
	    | _ -> typ, false
	  end
  in
    (* printf "<EXPAND:AFTER>@[%a@]@\n" print_typ rettyp; *)
    rettyp
and find_template typs =
  List.fold_left
    (fun typopt typ ->
       match typopt, typ with
	 | Some _, _ -> typopt
	 | None, TVar tv -> None
	 | None, TTypPlus typs -> find_template typs
	 | None, _ -> Some typ)
    None typs
and expand_until_saturate loc old_tmaxid cs =
  let cs_r = ref cs in begin
      try
      while true do
	(* printf "<LOOP>@[%a@]@\n" print_cs !cs_r; *)
	let cs, dirty = expand_tvar_and_flatten loc old_tmaxid !cs_r in
	let cs = normalize_cs cs in
	  cs_r := cs;
	  if not dirty then raise ExpansionSaturated
      done
      with
	| ExpansionSaturated -> ()
    end;
    !cs_r

let rec subst_ovar_typ ov o typ =
  match typ with
    | TRef(typ, o', Some w') -> TRef(subst_ovar_typ ov o typ, subst_ovar_owner ov o o', Some (subst_ovar_owner ov o w'))
    | TRef(typ, o', None) -> TRef(subst_ovar_typ ov o typ, subst_ovar_owner ov o o', None)
    | TRecord fts -> TRecord 
	(List.rev_map 
	   (function (fn, typ, Some o') -> (fn, subst_ovar_typ ov o typ, Some (subst_ovar_owner ov o o'))
	      | (fn, typ, None) -> (fn, subst_ovar_typ ov o typ, None)) fts)
    | TFun(typs1, te1, typs2, te2) ->
	TFun(List.map (subst_ovar_typ ov o) typs1, subst_ovar_tenv ov o te1,
	     List.map (subst_ovar_typ ov o) typs2, subst_ovar_tenv ov o te2)
    | TVarargs (* | TPlus _ *) | TVar _ | TSomething -> typ
    | TTypPlus typs -> TTypPlus (List.rev_map (subst_ovar_typ ov o) typs)
and subst_ovar_tenv ov o te =
  List.rev_map (fun (vi, typ) -> (vi, subst_ovar_typ ov o typ)) te

(*
let rec apply_typ funtyp preargs postargs =
  match funtyp with
    | TFun(preargs', postargs')
*)

let typ_deref loc rw typpre typpost =
  let (typpre', fpre, wpre), (typpost', fpost, wpost) = destruct_reftyp loc typpre, destruct_reftyp loc typpost in
    match rw, wpre, wpost with
      | `Read, Some wpre, Some wpost -> 
	  typpre', typpost', merge_cs [lt_owner loc owner_zero wpre; eq_owner loc wpre wpost; eq_owner loc fpre fpost]
      | `Read, None, None ->
	  typpre', typpost', merge_cs [eq_owner loc fpre fpost]
      | `Write, Some wpre, Some wpost -> 
	  typpre', typpost', merge_cs [eq_owner loc wpre owner_one; eq_owner loc wpre wpost; eq_owner loc fpre fpost]
      | `Write, None, None ->
	  typpre', typpost', merge_cs [eq_owner loc fpre fpost]
      | _ ->
	fstwarn "<typpre>@[%a@]@\n<typpost>@[%a@]@\n" print_typ typpre print_typ typpost;
	failwith "typ_deref: type error"

let rec make_unconstrained_te_lv lv te =
  let rec get_base_var lv =
    match lv with
      | Var x, _ -> x
      | Mem(e), _ -> get_base_var_e e
(*
      | _ ->
	  fstwarn "<lv>@[%a@]@\n" print_lval lv;
	  failwith "get_base_var: Not implemented"
*)
  and get_base_var_e e =
    match e with
      | Lval lv -> get_base_var lv
      | _ ->
	  fstwarn "<e>@[%a@]@\n" print_exp e;
	  failwith "get_base_var_e: Not implemented"
  in
  let x = get_base_var lv in
  let te, terest = List.partition (fun (x', _) -> x'.vname = x.vname) te in
  let te = List.rev_map (fun (x, typ) -> x, make_unconstrained_typ lv typ) te in
    te @ terest
and make_unconstrained_typ lv typ =
  let rec track_ofs ofs typ =
    (* printf "track_ofs: typ: @[%a@], ofs: @[%a@]@\n%!" print_typ typ print_offset ofs; *)
    let typ = find_typ typ in
    match typ, ofs with
      | _, NoOffset -> 
	  let typ' = rename_typ ~rename_owner_fun:rename_owner_renamefrac typ in
(*
	    if !FreeSafeTyFlag.debug then
	      (fstwarn "renamed.<lv>@[%a@]@\n<typ>@[%a@]@\n<typ'>@[%a@]@\n" print_lval lv print_typ typ print_typ typ';
	       print_flush ());
*)
	    typ'
      | TRecord fts, Field(f, ofs) ->
	  (* [XXX] Is this OK? If fn = f.name, returning new_ovar() instaed of o seems to be appropreate *)
	  TRecord (List.rev_map (fun (fn, typ', o) -> if fn = f.fname then fn, track_ofs ofs typ', o else fn, typ', o) fts)
      | TVar tv, Field(_, _) ->
	  (* [XXX] Presumably not correct. Run with --dumpIntermFile
	     --measureTime --outputLinearProg --printInferredTypes
	     test/small1/cdrom.tmp.c *)
	  let ciltyp' = 
	    if isVoidPtrType (snd tv.id) then failwith "track_ofs: void* case: not implemented"
	    else 
	      match unrollType (snd tv.id) with (TComp _) as t' -> t' | _ -> failwith "track_ofs: record: type error"
	  in
	  let typ' = typ_of_ciltyp ~unfold:true ciltyp' in
	    
	  (* [XXX] currentLoc *)
	  let _, b = union_typ !currentLoc tv typ' in
	    assert(b);
	    track_ofs ofs typ'
      | _ ->
	  fstwarn "<typ>@[%a@]@\n<ofs>@[%a@]@\n" print_typ typ print_offset ofs;
	  failwith "track_ofs: Not implemented"
  in
    match typ, lv with
      | _, (Var x, ofs) ->
	  track_ofs ofs typ
      | TRef(typ', f, w), (Mem(Lval(lv')), ofs) ->
	  TRef(make_unconstrained_typ (add_ofs lv' ofs) typ', f, w)
      | _ ->
	  fstwarn "<lv>@[%a@]<typ>@[%a@]@\n" print_lval lv print_typ typ;
	  failwith "make_unconstrained_typ: Not implemented."
    
let subst_ovar_constr ov o c =
  match c with
    | CLeOwner(o1, o2, loc) -> CLeOwner(subst_ovar_owner ov o o1, subst_ovar_owner ov o o2, loc)
    | CEqOwner(o1, o2, loc) -> CEqOwner(subst_ovar_owner ov o o1, subst_ovar_owner ov o o2, loc)    
    | CSubtype(typ1, typ2, loc) -> CSubtype(subst_ovar_typ ov o typ1, subst_ovar_typ ov o typ2, loc)
    | CEmpty(tv, _) -> c
    | CWfTyp(typ, loc) -> CWfTyp(subst_ovar_typ ov o typ, loc)

(*
let subst_ovar_cs ov o cs =
  (* [XXX] apply uniq to the following expression *)
  List.rev_map (fun c -> subst_ovar_constr ov o c) cs
*)

(*
let rec make_template_tenv file =
  List.fold_left
    (fun te g ->
       match g with
	 | GVarDecl(vi, _) | GVar(vi, _, _) ->
	     if mem_tenv vi te then te else (vi, typ_of_ciltyp vi.vtype)::te
	 | GFun(fd, _) ->
	     (* if mem_tenv fd.svar te then te else (fd.svar, typ_of_ciltyp fd.svar.vtype)::te *)
	     (fd.svar, find_global_env fd.svar)::te
	 | GType _ | GCompTag _ | GCompTagDecl _ | GEnumTag _ | GEnumTagDecl _ | GAsm _
	 | GPragma _ | GText _ -> te)
    empty_tenv
    file.globals
*)

(**** For debugging ****)
let instr_h = Hashtbl.create 128
let stmt_h = Hashtbl.create 128
exception VarargsFound of tenv * tenv * typ list * typ list * constr_set

let dump_genconstr_debug prefix prete postte cs =
  fstdebug "<%s>@\n<tenv>@\n" prefix;
  List.iter2
    (fun (x1, typ1) (x2, typ2) ->
       assert(x1.vname = x2.vname);
       printf "%s:@[@[%a@]@\n-->@\n@[%a@]@]@\n" x1.vname print_typ typ1 print_typ typ2)
    (sort_tenv prete) (sort_tenv postte);
  fstdebug "<constr>@[%a@]@\n" print_cs cs

let dump_instr_hash () =
  Hashtbl.iter 
    (fun i (pre, post, cs) -> 
       (*
	 fstdebug "<instr>@[%a@]@\n<pre>@[%a@]@\n<post>@[%a@]@\n<cs>@[%a@]@\n"
	 print_instr i print_tenv pre 
	 print_tenv post print_cs cs*)
       (fstdebug "@[%a@]@\n" print_instr i;
	dump_genconstr_debug "instr"  pre post (normalize_cs cs))
    )
    instr_h

let rec proj_ovs_tenv ovs te =
  List.filter (fun (_, typ) -> not (OVS.is_empty (OVS.inter ovs (get_freeovids_typ typ)))) te
let rec proj_ovs_cs ovs cs =
  CS.filter (fun c -> not (OVS.is_empty (OVS.inter ovs (get_freeovids_constr c)))) cs
    
let dump_relevant_instr desc =
  match desc with
    | `Constr c ->
	Hashtbl.iter
	  (fun i (pre, post, cs) ->
	     CS.iter (fun c' -> 
			if c == c' then
			  dump_genconstr_debug "dump" pre post cs) cs)
	  instr_h
    | `Ovs ovs ->
	(* printf "passed ovs: @[%a@]@\n" print_ovs ovs; *)
	(*print_only := Some ovs;*)
	Hashtbl.iter
	  (fun i (pre, post, cs) ->
	     if OVS.is_empty
	       (OVS.inter ovs 
		  (OVS.union (OVS.union (get_freeovids_tenv pre) (get_freeovids_tenv post)) (get_freeovids_cs cs))) then
		 ()
	     else begin
	       printf "<instr>@[%a@]@\n" print_instr i;
	       (* dump_genconstr_debug "dump" (proj_ovs_tenv ovs pre) (proj_ovs_tenv ovs post) (proj_ovs_cs ovs cs) *)
	       printf "<pre>@[%a@]@\n" print_tenv (*(proj_ovs_tenv ovs pre)*)(normalize_tenv (get_instrLoc i) pre);
	       printf "<post>@[%a@]@\n" print_tenv (*(proj_ovs_tenv ovs post)*)(normalize_tenv (get_instrLoc i) post);
	       printf "<cs>@[%a@]@\n" print_cs (normalize_cs (*(proj_ovs_cs ovs cs)*)cs)
	     end)
	  instr_h;
	print_only := None

let dump_stmt_hash () =
  Hashtbl.iter
    (fun s (pre, post, cs) ->
       fstdebug "@[%a@]@\n" print_stmt s;
       dump_genconstr_debug "stmt" pre post (normalize_cs cs))
    stmt_h

let rec f_glob g =
  match g with
    | GFun(fd, loc) ->
	begin match !FreeSafeTyFlag.doOnly with
	  | None ->
	      let funtyp = find_global_env fd.svar in
		f_fundec loc funtyp fd
	  | Some l when List.mem fd.svar.vname l ->
	      let funtyp = find_global_env fd.svar in
		f_fundec loc funtyp fd
	  | _ -> empty_cs
	end
    | GType _ | GCompTag _ | GCompTagDecl _
    | GEnumTag _ | GEnumTagDecl _ | GVarDecl _
    | GVar _ | GAsm _ | GPragma _ | GText _ -> empty_cs
and f_fundec loc funtyp fd =
(*
  if fd.svar.vname = "register_cdrom" then
    FreeSafeTyFlag.debug := true;
*)
  (* fstwarn "Processing fundec %s@\n%!" fd.svar.vname; *)
  match funtyp with
    | TFun(preargs, preglobs, postargs, postglobs) ->
	let rec extract_laststmts fd = extract_b fd.sbody
	and extract_b b = List.fold_left (fun l s -> l @ (extract_s s)) [] b.bstmts
	and extract_s s =
	  match s.skind with
	    | Return _ -> [s]
	    | If(_, b1, b2, _) | TryFinally(b1, b2, _) | TryExcept(b1, _, b2, _) 
		-> (extract_b b1) @ (extract_b b2)
	    | Switch(_, b, _, _) | Loop(b, _, _, _) | Block b ->
		extract_b b
	    | _ -> []
	in
	let local_post_te = combine_tenv fd.slocals (List.map (fun vi -> typ_of_ciltyp vi.vtype) fd.slocals) in
	let formal_post_te = combine_tenv fd.sformals postargs in
	let postte = disjoint_union_tenv (disjoint_union_tenv postglobs local_post_te) formal_post_te in
	let laststmts = extract_laststmts fd in
	let prete, c1 = f_block loc laststmts postte fd.sbody in
	let formal_pre_te, local_pre_te, preglobs' =
	  List.fold_left
	    (fun (formal, local, glob) (vi, typ) ->
	       if vi.vglob then (formal, local, add_tenv (vi, typ) glob)
	       else if List.exists (fun vi' -> vi.vname = vi'.vname) fd.sformals then
		 (add_tenv (vi, typ) formal, local, glob)
	       else (formal, add_tenv (vi, typ) local, glob))
	    (empty_tenv, empty_tenv, empty_tenv)
	    prete
	in
	let c2 = sub_tenv loc (combine_tenv fd.sformals preargs) formal_pre_te in
	let c3 = sub_tenv loc preglobs preglobs' in
	let c4 = empty_cs_tenv loc local_post_te in
	let c5 = empty_cs_tenv loc local_pre_te in
	let c6 = if fd.svar.vname = "main" then empty_cs_tenv loc postglobs else empty_cs in
	let cs = merge_cs [c1; c2; c3; c4; c5; c6] in
	  (*
	    if !FreeSafeTyFlag.debug then
	    fstdebug "dec of @[%s@]@\nc1:@[%a@]c2:@[%a@]@\nc3@[%a@]@\nc4@[%a@]@\nc5:@[%a@]@\nc6:@[%a@]@\n" 
	    fd.svar.vname print_cs c1 print_cs c2 print_cs c3 print_cs c4 print_cs c5 print_cs c6;
	  *)
	  cs
    | TVarargs | TTypPlus _ (* | TPlus _ *) | TVar _ | TRecord _ | TSomething | TRef _ -> failwith "f_fundec: type mismatch"
and f_block loc laststmts postenv b =
  let initial_s = match b.bstmts with [] -> failwith "f_block: empty block" | hd::_ -> hd in
  let rec helper tobe processed c =
    match tobe with
      | [] -> c, processed
      | (s, postte)::tl ->
	  (* fstwarn "f_block: helper: processing @[%a@]@\n%!" print_stmt s; *)
	  (* if !FreeSafeTyFlag.debug then printf "OK4!@\n%!"; *)
	  let prete, c1 = f_stmt postte s in
	    (* if !FreeSafeTyFlag.debug then printf "OK5!@\n%!"; *)
	    let _ = assert(not (List.exists (fun (_, _, s') -> s == s') processed)) in
	  (* let _ = fstwarn "f_block: added to processed:@\n@[%a@]@\n" print_stmt s in *)
	  let new_processed = (prete, postte, s)::processed in
	  let new_tobe, c2 = 
	    List.fold_left
	      (fun (new_tobe, c) s' ->
(*
		 if !FreeSafeTyFlag.debug then
		   printf "Processing @[%a@]@\n%!" print_stmt s';
*)
		 try
		   let _, postte, _ = List.find (fun (_, _, s'') -> s' == s'') new_processed in
		   let c' = sub_tenv loc postte prete in
(*
		     if !FreeSafeTyFlag.debug then
		       (fstwarn "f_block: helper<1>:@\n<from stmt>@[%a@]@\n<to stmt>@[%a@]@\n"
			  print_stmt s' print_stmt s;
			dump_genconstr_debug "block" postte prete c';
			print_flush ());
*)
		     new_tobe, (union_cs c c')
		 with
		   | Not_found -> 
(*
		       if !FreeSafeTyFlag.debug then
			 (fstwarn "f_block: helper<2>:@\n<from stmt>@[%a@]@\n<to stmt>@[%a@]@\n"
			    print_stmt s' print_stmt s;
			  dump_genconstr_debug "block" postte prete empty_cs;
			  print_flush ());
*)
		       if List.exists (fun (s, _) -> s == s') new_tobe then begin
			 (* if !FreeSafeTyFlag.debug then printf "OK1!@\n%!"; *)
			 new_tobe, c
		       end else begin
(*
			 if !FreeSafeTyFlag.debug then
			   printf "OK1!@\n%!";
*)
			 (s', prete)::new_tobe, c
		       end)
	      (tl, empty_cs)
	      s.preds
	  in
	    helper new_tobe new_processed (merge_cs [c; c1; c2])
  in
  let initial_tobe = List.rev_map (fun s -> s, postenv) laststmts in
  let cs, processed = helper initial_tobe [] empty_cs in
  let prete =
    try
      let (pre, _, _) = List.find (fun (_, _, s') -> s' == initial_s) processed in
	pre
    with
      | Not_found -> failwith "f_block: initial statement is not processed."
  in
    (*
      let prete =
      try
      let entry = List.find (fun (b', _) -> b == b') !FreeSafeTyCheckNullLvals.nulllvals in
      let lvals = snd entry in
      List.fold_left (fun te lv -> make_unconstrained_te_lv lv te) prete lvals
      with
      | Not_found -> prete
      in
    *)
(*
    if !FreeSafeTyFlag.debug then
      (fstdebug "<f_block>@\n@[%a@]@\n<prete>@\n@[%a@]@\n<cs>@\n@[%a@]@\n" print_block b print_tenv prete print_cs cs;
       dump_genconstr_debug "f_block" prete postenv cs);
*)
    prete, cs
and f_stmt postte s = (* failwith "f_stmt: not implemented yet." *)
  (* let _ = if !FreeSafeTyFlag.debug then fstdebug "Processing stmt @[%a@]@\n%!" print_stmt s in *)
  let prete, cs =
    match s.skind with
      | Instr is -> f_instrs postte is
      | Return _ | Goto _ | Break _ | Continue _
      | If _ | Switch _ | Loop _ | Block _
      | TryFinally _ | TryExcept _ -> postte, empty_cs
  in
    (*
      let prete =
      try
      let entry = List.find (fun (s', _) -> s == s') !FreeSafeTyCheckNullLvals.nulllvals in
      let lvals = snd entry in
      List.fold_left (fun te lv -> make_unconstrained_te_lv lv te) prete lvals
      with
      | Not_found -> prete
      in
    *)
(*
    if !FreeSafeTyFlag.debug then
      (* fstdebug "<f_stmt>@\n@[%a@]@\n<prete>@\n@[%a@]@\n<postte>@\n@[%a@]@\n<constr>@\n@[%a@]@\n" *)
      dump_genconstr_debug "f_stmt" prete postte cs;
*)
    Hashtbl.add stmt_h s (prete, postte, cs);
    prete, cs
and f_instrs postte is =
  match is with
    | [] -> postte, empty_cs
    | hd::tl ->
	let postte', c1 = f_instrs postte tl in
	let prete, c2 = f_instr postte' hd in
	  prete, merge_cs [c1; c2]
and gen_pre_te loc post post' pre' =
  let post, post', pre' = sort_tenv post, sort_tenv post', sort_tenv pre' in
  let c_ref = ref empty_cs in

  (* Generate a pre-type-environment pre and a constraint set by the
     following rules.  (i) If post'(x) is not defined, pre(x) =
     post(x).  (ii) If post'(x) and pre'(x) are defined, pre(x) =
     pre'(x) + te and post'(x) + te = post(x) where te is a renaming
     of post'(x) (which is also a renaming of pre'(x)).  *)
  let rec helper post post' pre' =
    match post, post', pre' with
      | _, [], [] -> post
      | (v1, typ1)::tl1, (v1', typ1')::tl1', (v2', typ2')::tl2' when v1'.vname = v2'.vname ->
	  let cmp = compare v1.vname v1'.vname in
	    if cmp = 0 then 
	      (c_ref := union_cs !c_ref (eq_typ loc typ1' typ1);
	       (v2', typ2')::(helper tl1 tl1' tl2'))
	    else if cmp < 0 then (v1, typ1)::(helper tl1 post' pre')
	    else
	      begin 
		match typ1' with 
		    TFun _ when v1'.vglob ->
		      helper post tl1' tl2'
		  | _ ->
		      fstwarn "<post>@[%a@]@\n<post'>@[%a@]@\n<pre'>@[%a@]@\n"
			print_tenv post print_tenv post' print_tenv pre';
		      failwith "gen_pre_te: helper: a variable in post' is not contained in post"
	      end
      | _ -> 
	  fstdebug "<post>@[%a@]@\n<post'>@[%a@]@\n<pre'>@[%a@]@\n"
	    print_tenv post print_tenv post' print_tenv pre';
	  failwith "gen_pre_te: helper: post' does not match with pre'."
  in
    !c_ref, helper post post' pre'
and f_instr postte i =
  (* let _ = if !FreeSafeTyFlag.debug then fstdebug "Processing instr @[%a@]@\n%!" print_instr i in *)
  let prete, cs =
    match i with
	(*
	  | Set((Mem ep, NoOffset), e, _) ->
	  let pre_te_ep, post_te_ep, pre_t_ep, post_t_ep, c1 = f_e loc `Read ep in
	  let pre_te_e, post_te_e, pre_t_e, post_t_e, c2 = f_e loc `Read e in
	  let pre_t_ep', f_e locp_pre, w_ep_pre = destruct_reftyp pre_t_ep in
	  let post_t_ep', f_e locp_post, w_ep_post = destruct_reftyp post_t_ep in
	(*
	  let passed_te = rename_tenv (tensor_tenv loc post_te_ep post_te_e) in
	  let post_te' = tensor_tenv loc passed_te (tensor_tenv loc post_te_ep post_te_e) in
	  let pre_te' = tensor_tenv loc passed_te (tensor_tenv loc pre_te_ep pre_te_e) in
	*)
	  let post_te' = (tensor_tenv loc post_te_ep post_te_e) in
	  let pre_te' = (tensor_tenv loc pre_te_ep pre_te_e) in
	  let c3 = merge_cs [wf_te post_te'; wf_te pre_te'] in
	  let c4 = merge_cs 
	  [empty_typ loc pre_t_ep'; sub_type loc pre_t_e (tensor_typ loc post_t_ep' post_t_e);
	  le_owner f_e locp_post f_e locp_pre; le_owner f_e locp_pre f_e locp_post;
	  le_owner (OFrac 1.0) w_ep_pre; le_owner (OFrac 1.0) w_ep_post] 
	  in
	  let c5, pre_te = gen_pre_te postte post_te' pre_te' in
	  pre_te, merge_cs [c1; c2; c3; c4; c5]
	  | Set((Var x, NoOffset), e, _) ->
	  let pre_te_e, post_te_e, pre_t_e, post_t_e, c1 = f_e loc `Read e in
	  let pre_t_x, post_t_x = typ_of_ciltyp x.vtype, typ_of_ciltyp x.vtype in
	(*
	  let passed_te = rename_tenv (tensor_tenv loc [x, post_t_x] post_te_e) in
	  let pre_te' = tensor_tenv loc passed_te (tensor_tenv loc [x, pre_t_x] pre_te_e) in
	  let post_te' = tensor_tenv loc passed_te (tensor_tenv loc [x, post_t_x] post_te_e) in
	*)
	  let pre_te' = (tensor_tenv loc [x, pre_t_x] pre_te_e) in
	  let post_te' = (tensor_tenv loc [x, post_t_x] post_te_e) in
	  let c2 = merge_cs 
	  [empty_typ loc pre_t_x; sub_type loc pre_t_e (tensor_typ loc post_t_x post_t_e); wf_te pre_te'; wf_te post_te']
	  in
	  let c3, pre_te = gen_pre_te loc postte post_te' pre_te' in
	  pre_te, merge_cs [c1; c2; c3]
	*)
      | Set(lv, e, loc) ->
	  let pre_te_lv, post_te_lv, pre_t_lv, post_t_lv, c1 = f_lv loc `Write lv in
	  let pre_te_e, post_te_e, pre_t_e, post_t_e, c2 = f_e loc `Read e in
	    (* let passed_te = rename_tenv (tensor_tenv loc pre_te_lv pre_te_e) in *)
	  let pre_te' = (tensor_tenv loc pre_te_lv pre_te_e) in
	  let post_te' = (tensor_tenv loc post_te_lv post_te_e) in
	  let posttyp' = tensor_typ loc post_t_lv post_t_e in
	  let c3 = merge_cs [wf_te loc pre_te'; wf_te loc post_te';
			     empty_typ loc pre_t_lv; eq_typ loc pre_t_e posttyp';
			     wf_typ loc posttyp'] in
	  let c4, pre_te = gen_pre_te loc postte post_te' pre_te' in
	    pre_te, merge_cs [c1; c2; c3; c4]
      | Call(None, Lval(Var vi, NoOffset), [e_ret; e_size], loc) when vi.vname = "malloc" ->
	  let pre_te_e_size, post_te_e_size, pre_t_e_size, post_t_e_size, c1 = f_e loc `Read e_size in
	  let pre_te_e_ret, post_te_e_ret, pre_t_e_ret, post_t_e_ret, c2 = f_e loc `Write e_ret in
	  let post_te' = (tensor_tenv loc post_te_e_ret post_te_e_size) in
	  let pre_te' = (tensor_tenv loc pre_te_e_ret pre_te_e_size) in
	  let c3 = merge_cs [wf_te loc post_te'; wf_te loc pre_te'] in
	  let post_t_e_ret', post_f1, post_w1 = destruct_reftyp loc post_t_e_ret in
	  let post_t_e_ret'', post_f2, post_w2 = destruct_reftyp loc post_t_e_ret' in
	  let pre_t_e_ret', pre_f1, pre_w1 = destruct_reftyp loc pre_t_e_ret in
	  let pre_t_e_ret'', pre_f2, pre_w2 = destruct_reftyp loc pre_t_e_ret' in
	  let c4write =
	    match pre_w1, pre_w2, post_w1, post_w2 with
	      | None, None, None, None -> empty_cs
	      | Some pre_w1, Some pre_w2, Some post_w1, Some post_w2 ->
		  merge_cs 
		    [eq_owner loc (OFrac 1.0) pre_w1; eq_owner loc pre_w2 (OFrac 0.0); 
		     eq_owner loc (OFrac 1.0) post_w1; eq_owner loc (OFrac 1.0) post_w2]
	      | _ -> failwith "Call: malloc: type error"
	  in
	  let c4 = merge_cs 
	    [eq_typ loc post_t_e_size TSomething; eq_typ loc pre_t_e_size TSomething;
	     empty_typ loc pre_t_e_ret';
	     eq_owner loc pre_f2 (OFrac 0.0); eq_owner loc (OFrac 1.0) post_f2;
	     eq_owner loc pre_f1 post_f1; empty_typ loc pre_t_e_ret''; empty_typ loc post_t_e_ret''; c4write] 
	  in
	    let c5, pre_te = gen_pre_te loc postte post_te' pre_te' in
	      pre_te, merge_cs [c1; c2; c3; c4; c5]
      | Call(None, Lval(Var vi, NoOffset), [e_ret; e_size; _], loc) when vi.vname = "kmalloc" || vi.vname = "kzalloc" ->
	  let pre_te_e_size, post_te_e_size, pre_t_e_size, post_t_e_size, c1 = f_e loc `Read e_size in
	  let pre_te_e_ret, post_te_e_ret, pre_t_e_ret, post_t_e_ret, c2 = f_e loc `Write e_ret in
	  let post_te' = (tensor_tenv loc post_te_e_ret post_te_e_size) in
	  let pre_te' = (tensor_tenv loc pre_te_e_ret pre_te_e_size) in
	  let c3 = merge_cs [wf_te loc post_te'; wf_te loc pre_te'] in
	  let post_t_e_ret', post_f1, post_w1 = destruct_reftyp loc post_t_e_ret in
	  let post_t_e_ret'', post_f2, post_w2 = destruct_reftyp loc post_t_e_ret' in
	  let pre_t_e_ret', pre_f1, pre_w1 = destruct_reftyp loc pre_t_e_ret in
	  let pre_t_e_ret'', pre_f2, pre_w2 = destruct_reftyp loc pre_t_e_ret' in
	  let c4write =
	    match pre_w1, pre_w2, post_w1, post_w2 with
		None, None, None, None -> empty_cs
	      | Some pre_w1, Some pre_w2, Some post_w1, Some post_w2 ->
		  merge_cs
		    [eq_owner loc (OFrac 1.0) pre_w1; eq_owner loc pre_w2 (OFrac 0.0); 
		     eq_owner loc (OFrac 1.0) post_w1; eq_owner loc (OFrac 1.0) post_w2]
	      | _ -> failwith "Call: kmalloc: type error"
	  in
	  let c4 = merge_cs 
	    [eq_typ loc post_t_e_size TSomething; eq_typ loc pre_t_e_size TSomething;
	     empty_typ loc pre_t_e_ret';
	     c4write;
	     eq_owner loc pre_f2 (OFrac 0.0); eq_owner loc (OFrac 1.0) post_f2;
	     eq_owner loc pre_f1 post_f1; empty_typ loc pre_t_e_ret''; empty_typ loc post_t_e_ret''] 
	  in
	  let c5, pre_te = gen_pre_te loc postte post_te' pre_te' in
	      pre_te, merge_cs [c1; c2; c3; c4; c5]
      | Call(None, Lval(Var vi, NoOffset), [e_ptr], loc) when vi.vname = "free" || vi.vname = "xfree" || vi.vname = "kfree" ->
	  let pre_te_e_ptr, post_te_e_ptr, pre_t_e_ptr, post_t_e_ptr, c1 = f_e loc `Read e_ptr in
	    (*
	      let passed_te = rename_tenv post_te_e_ptr in
	      let post' = tensor_tenv loc post_te_e_ptr passed_te in
	      let pre' = tensor_tenv loc pre_te_e_ptr passed_te in
	    *)
	  let post' = post_te_e_ptr in
	  let pre' = pre_te_e_ptr in
	    (* let c2 = merge_cs [wf_te loc post'; wf_te loc pre'] in *)
	  let post_t_e_ptr', post_f, post_w = destruct_reftyp loc post_t_e_ptr in
	  let pre_t_e_ptr', pre_f, pre_w = destruct_reftyp loc pre_t_e_ptr in
	  let c3 = 
	    match pre_w, post_w with
		Some pre_w, Some post_w ->
		  merge_cs
		    [eq_owner loc (OFrac 1.0) pre_f; eq_owner loc post_f (OFrac 0.0);
		     eq_owner loc (OFrac 1.0) pre_w; eq_owner loc post_w (OFrac 0.0);
		     empty_typ loc post_t_e_ptr'; empty_typ loc pre_t_e_ptr']
	      | None, None ->
		  merge_cs
		    [eq_owner loc (OFrac 1.0) pre_f; eq_owner loc post_f (OFrac 0.0);
		     empty_typ loc post_t_e_ptr'; empty_typ loc pre_t_e_ptr']
	      | _ -> failwith "Call: free: type error."
	  in
	  let c4, pre_te = gen_pre_te loc postte post' pre' in
	    pre_te, merge_cs [c1; (*c2;*) c3; c4]
      | Call(None, Lval(Var f, NoOffset), [e_ptr], loc) when f.vname = "_fst_assume_null" ->
	  let lv = stripCastChain e_ptr in
	  let pre = make_unconstrained_te_lv lv postte in
(*
	    fstdebug "<instr>@[%a@]@\n" print_instr i;
	    dump_genconstr_debug "<assume_null>" pre postte empty_cs;
*)
	    pre, empty_cs
      | Call(None, Lval(Var f, NoOffset), [e_p1; e_p2], loc) when f.vname = "_fst_assert_eq" || f.vname = "_fst_assert_sameregion" ->
	  let pre_te_p1, post_te_p1, pre_typ_p1, post_typ_p1, c1 = f_e loc `Read e_p1 in
	  let pre_te_p2, post_te_p2, pre_typ_p2, post_typ_p2, c2 = f_e loc `Read e_p2 in
	  let post_te', pre_te' = (tensor_tenv loc post_te_p1 post_te_p2), (tensor_tenv loc pre_te_p1 pre_te_p2) in
	  let c3, pre_te = gen_pre_te loc postte post_te' pre_te' in
	  let pre_typ, post_typ = tensor_typ loc pre_typ_p1 pre_typ_p2, tensor_typ loc post_typ_p1 post_typ_p2 in
	  let c4 = merge_cs [eq_typ loc pre_typ post_typ;
			     wf_te loc post_te'; wf_te loc pre_te'; wf_typ loc pre_typ; wf_typ loc post_typ] in
	    pre_te, merge_cs [c1; c2; c3; c4]
      | Call(None, Lval(Var f, NoOffset), [e_p], loc) when f.vname = "_fst_assert_empty" ->
	  let pre_te, post_te, pre_typ, post_typ, c1 = f_e loc `Read e_p in
	  let c2, prete = gen_pre_te loc postte post_te pre_te in
	  let c3 = merge_cs [eq_typ loc pre_typ post_typ;
			     empty_typ loc pre_typ; empty_typ loc post_typ]
	  in
	    prete, merge_cs [c1; c2; c3]
      | Call(None, Lval(Var f, NoOffset), [e_p; e_frac], loc) when f.vname = "_fst_assert_writecap" ->
	  begin match e_frac with
	    | Const(CReal(f, _, _)) ->
		let pre_te_p, post_te_p, pre_typ_p, post_typ_p, c1 = f_e loc `Read e_p in
		let c2, prete = gen_pre_te loc postte post_te_p pre_te_p in
		let _, _, w = destruct_reftyp loc pre_typ_p in
		let c3 = 
		  match w with Some w -> merge_cs [eq_typ loc pre_typ_p post_typ_p; eq_owner loc w (OFrac f)] 
		    | None -> eq_typ loc pre_typ_p post_typ_p
		in
		  prete, merge_cs [c1; c2; c3]
	    | _ -> failwith "Type error in _fst_assert_writecap"
	  end
      | Call(None, Lval(Var f, NoOffset), [e_p; e_frac], loc) when f.vname = "_fst_assert_freeob" ->
	  if !FreeSafeTyFlag.assertFreeob then
	    begin match e_frac with
	      | Const(CReal(f, _, _)) ->
		  let pre_te_p, post_te_p, pre_typ_p, post_typ_p, c1 = f_e loc `Read e_p in
		  let c2, prete = gen_pre_te loc postte post_te_p pre_te_p in
		  let _, free, _ = destruct_reftyp loc pre_typ_p in
		  let c3 = merge_cs [eq_typ loc pre_typ_p post_typ_p; eq_owner loc free (OFrac f)] in
		    prete, merge_cs [c1; c2; c3]
	      | _ -> failwith "Type error in _fst_assert_freeob"
	    end
	  else
	    postte, empty_cs
      | Call(None, (*Lval(Var f, NoOffset)*)e_f, eargs, loc) ->
	  let f_typ, te_f, c0 = 
	    match e_f with
	      | Lval(Var f, NoOffset) -> 
		  find_global_env f, empty_tenv, empty_cs
	      | _ -> 
		  let te_pre, te_post, typ_pre, typ_post, c1 = f_e loc `Read e_f in
		    typ_pre, te_pre, merge_cs [c1; eq_typ loc typ_pre typ_post; eq_tenv loc te_pre te_post]
	  in
	  (* let _ = fstwarn "Processing Call: @[%s@]@\n@[%a@]@\n%!" f.vname print_typ f_typ in *)
	  let preglob, postglob = get_glob_funtyp f_typ in
	  let make_te_t_c () = 
	    try
	      List.fold_right
		(fun e (tepre, tepost, targspre, targspost, c) ->
		   let tepre', tepost', tpre', tpost', c' = f_e loc `Read e in
		     match tpre' with
		       | TVarargs -> raise (VarargsFound(tepre, tepost, targspre, targspost, c))
		       | _ ->
			   tensor_tenv loc tepre tepre', tensor_tenv loc tepost tepost', tpre'::targspre, tpost'::targspost, union_cs c c')
		eargs ([], [], [], [], empty_cs)
	    with
	      | VarargsFound(prete, postte, prets, postts, cs) -> prete, postte, prets, postts, cs
	  in
	  let pre_te_eargs, post_te_eargs, pre_t_eargs, post_t_eargs, c1 = make_te_t_c () in
(*
	  let _ = fstdebug "<pre_te_eargs>@[%a@]@\n<post_te_eargs>@[%a@]@\n<pre_t_eargs>@[%a@]@\n<post_t_eargs>@[%a@]@\n<c1>@[%a@]@\n"
	    print_tenv pre_te_eargs print_tenv post_te_eargs (print_list print_typ ~sep:",") pre_t_eargs
	    (print_list print_typ ~sep:",") post_t_eargs print_cs c1
	  in
*)
	    (* let post_te_eargs, post_t_eargs, c2 = make_te_t_c () in *)
	    (* let rest_te_eargs, rest_t_eargs, c3 = make_te_t_c () in *)

(*
	  let rest_te_eargs, c2 = rename_tenv post_te_eargs, empty_cs in
	  let post_te' = tensor_tenv loc (tensor_tenv loc post_te_eargs rest_te_eargs) postglob in
	  let pre_te' = tensor_tenv loc (tensor_tenv loc pre_te_eargs rest_te_eargs) preglob in
*)
	  let c2 = empty_cs in
	  let post_te' = tensor_tenv loc (tensor_tenv loc post_te_eargs  postglob) te_f in
	  let pre_te' = tensor_tenv loc (tensor_tenv loc pre_te_eargs preglob) te_f in
	  let c3 = apply_typ loc f_typ pre_t_eargs post_t_eargs in
	  let c4, pre_te = gen_pre_te loc postte post_te' pre_te' in
	  let c5 = merge_cs [wf_te loc post_te'; wf_te loc pre_te'] in
(*
	  let _ = 
	    fstwarn "|c1|=@[%d@], |c2|=@[%d@], |c3|=@[%d@], |c4|=@[%d@], |c5|=@[%d@]@\n%!"
	      (CS.cardinal c1) (CSg.cardinal c2) (CS.cardinal c3) (CS.cardinal c4) (CS.cardinal c5)
	  in
*)
	    pre_te, merge_cs [c0; c1; c2; c3; c4; c5]
      | Asm _ ->
	  if !FreeSafeTyFlag.debug then
	    fstwarn "f_instr: <asm>: @[%a@]@\n" print_instr i;
	  postte, empty_cs
      | _ -> 
	  fstwarn "f_instr: Not implemented yet@\n@[%a@]@\n" print_instr i; 
	  (* postte, empty_cs *)
	  failwith "Not implemented yet"
  in
  let loc = get_instrLoc i in
  let prete, postte, cs = normalize_tenv loc prete, normalize_tenv loc postte, normalize_cs (*(expand_until_saturate (get_instrLoc i) !tmaxid cs)*) cs in
  (* let prete, postte, cs = normalize_tenv loc prete, normalize_tenv loc postte, cs in *)
    if !FreeSafeTyFlag.debug then
      (* dump_genconstr_debug "f_instr" prete postte cs; *) ();
    Hashtbl.add instr_h i (prete, postte, cs);
    prete, cs
and f_lv loc rw lv =
  let h, ofs = lv in
  let h_tepre, h_tepost, h_typpre, h_typpost, c1 =
    match h with
      | Var x -> 
	  let typ_xpre, typ_xpost = typ_of_ciltyp x.vtype, typ_of_ciltyp x.vtype in
	    [x, typ_xpre], [x, typ_xpost], typ_xpre, typ_xpost, empty_cs
      | Mem e ->
	  let te_epre, te_epost, typ_epre, typ_epost, c1 = f_e loc `Read e in
	    
	  (* If ofs <> NoOffset, e is being read.
	     If ofs = NoOffset and rw = `Write, e is being written. *)
	  let typ_epre', typ_epost', c2 = 
	    if ofs = NoOffset then typ_deref loc rw typ_epre typ_epost  
	    else typ_deref loc `Read typ_epre typ_epost
	  in
	    te_epre, te_epost, typ_epre', typ_epost', merge_cs [c1; c2]
  in
  let rec helper typpre typpost ofs =
    match ofs with
      | NoOffset -> empty_tenv, empty_tenv, typpre, typpost, empty_cs
      | Field(fi, ofs') ->
	  begin
	    let typpre, typpost = find_typ typpre, find_typ typpost in
	      match typpre, typpost with
	      | TRecord ftspre, TRecord ftspost -> 
		  let ftspre, ftspost = sort_fts ftspre, sort_fts ftspost in
		  let (typpre', o_pre), (typpost', o_post) = get_field fi.fname ftspre, get_field fi.fname ftspost in
		  let c1 =
		    (* [XXX] *)
		    match ofs', rw, o_pre, o_post with 
			(* If ofs' = NoOffset and rw = `Write, then typ.fi is being written.
			   In that case, write capability of fi in ftspre/ftspost has to be 1.
			   Otherwise (ofs' <> NoOffset or rw = `Read), fi is being read,
			   so that the write capability of fi should be greater than 0. *)
		      | _, _, None, None -> empty_cs
		      | NoOffset, `Write, Some o_pre, Some o_post -> 
			  merge_cs [eq_owner loc owner_one o_pre; eq_owner loc owner_one o_post;
				    eq_owner loc o_pre o_post]
		      | _, _, Some o_pre, Some o_post -> 
			  merge_cs [lt_owner loc owner_zero o_pre; lt_owner loc owner_zero o_post;
				    eq_owner loc o_pre o_post]
		      | _ -> failwith "f_lv: record: type error"
		  in
		  let tepre, tepost, typpre_res, typpost_res, c2 = helper typpre' typpost' ofs' in
		  let c3 = List.fold_left2 
		    (fun c (fpre, tpre, o_pre) (fpost, tpost, o_post) ->
		       assert(fpre = fpost);
		       if fpre = fi.fname then c 
		       else 
			 match o_pre, o_post with
			   | Some o_pre, Some o_post ->
			       merge_cs [c; eq_typ loc tpre tpost;
					 eq_owner loc o_pre o_post]
			   | None, None ->
			       merge_cs [c; eq_typ loc tpre tpost]
			   | _ -> failwith "f_lv: record write type: type error")
		    empty_cs ftspre ftspost
		  in
		    tepre, tepost, typpre_res, typpost_res, merge_cs [c1; c2; c3]
	      | TVar tvpre, TVar tvpost ->
		  let typpre_body, typpost_body = 
		    typ_of_ciltyp ~unfold:true (snd tvpre.id), 
		    typ_of_ciltyp ~unfold:true (snd tvpost.id) 
		  in
		  (* let _ = if !FreeSafeTyFlag.debug then printf "OK7!@[%a@]@\n%!" print_lval lv in *)
		  let c1 = merge_cs [eq_typ loc (TVar tvpre) typpre_body; eq_typ loc (TVar tvpost) typpost_body] in
		  let tepre, tepost, typpre, typpost, c2 = helper typpre_body typpost_body ofs in
		    tepre, tepost, typpre, typpost, merge_cs [c1; c2]
	      | _ ->
		  fstwarn "<typpre>@[%a@]@\n<typpost>@[%a@]@\n<ofs>@[%a@]@\n" print_typ typpre 
		    print_typ typpost print_offset ofs';
		  failwith "f_lv: helper: Field offset for non-record type."
	  end
      | Index(e, ofs') ->
	  let typpre', fpre, wpre = destruct_reftyp loc (find_typ typpre) in
	  let typpost', fpost, wpost = destruct_reftyp loc (find_typ typpost) in
	  let c1 =
	    (* [XXX] *)
	    match ofs', rw, wpre, wpost with 
		(* If ofs' = NoOffset and rw = `Write, then typ[e] is being written.
		   In that case, write capability of typ in has to be 1.
		   Otherwise (ofs' <> NoOffset or rw = `Read), fi is being read,
		   so that the write capability of fi should be greater than 0. *)
	      | _, _, None, None -> empty_cs
	      | NoOffset, `Write, Some wpre, Some wpost -> 
		  merge_cs [eq_owner loc owner_one wpre; eq_owner loc owner_one wpost;
			    eq_owner loc wpre wpost]
	      | _, _, Some wpre, Some wpost -> merge_cs [lt_owner loc owner_zero wpre; lt_owner loc owner_zero wpost]
	      | _ -> failwith "f_lv: array: type error"
	  in
	  let tepre, tepost, typpre_res, typpost_res, c2 = helper typpre' typpost' ofs' in
	  let te_e_pre, te_e_post, _, _, c3 = f_e loc `Read e in
	    (tensor_tenv loc tepre te_e_pre), (tensor_tenv loc tepost te_e_post), typpre_res, typpost_res, merge_cs [c1; c2; c3]
(*
      | ((TVar tvpre) as typpre'), ((TVar tvpost) as typpost') ->
	  let c1 = merge_cs [eq_typ loc (TVar tvpre) (typ_of_ciltyp (snd tvpre.id));
			     eq_typ loc (TVar tvpost) (typ_of_ciltyp (snd tvpost.id))] in
	  let tepre, tepost, typpre, typpost, c2 = helper typpre' typpost' ofs in
	    tepre, tepost, typpre, typpost, merge_cs [c1; c2]
*)
(*
      | _ -> 
	  fstwarn "<typpre>@[%a@]@\n<typpost>@[%a@]@\n<ofs>@[%a@]@\n" print_typ typpre
	    print_typ typpost print_offset ofs;
	  failwith "f_lv: helper: not implemented yet"
*)
  in
  let ofs_tepre, ofs_tepost, typpre, typpost, c2 = helper h_typpre h_typpost ofs in
  (* let _ = if !FreeSafeTyFlag.debug then printf "OK6!@[%a@]@\n%!" print_lval lv in *)
  let prete, postte = (tensor_tenv loc h_tepre ofs_tepre), (tensor_tenv loc h_tepost ofs_tepost) in
  let c3 = merge_cs [wf_te loc prete; wf_te loc postte] in
  let cs = normalize_cs (merge_cs [c1; c2; c3]) in
(*
  let _ =
    if !FreeSafeTyFlag.debug then begin
      fstdebug "<f_lv>@[%a@]@\n<typ>@[@[%a@]@\n-->@\n@[%a@]@]@\n" print_lval lv print_typ typpre print_typ typpost;
      dump_genconstr_debug "f_lv" prete postte cs
    end
  in
*)
    prete, postte, typpre, typpost, cs
and f_e loc rw e =
  match e with
    | Const (CStr _ | CWStr _) -> 
	empty_tenv, empty_tenv, 
	(TRef(TSomething, new_ovar(), None)), (TRef(TSomething, new_ovar(), None)), empty_cs
    | Const _ | SizeOf _ | SizeOfStr _ -> empty_tenv, empty_tenv, TSomething, TSomething, empty_cs
    | BinOp((Ge|Le|Gt|Lt|Mod|Div|Mult|BXor|PlusA|MinusA|Eq|Ne|BOr|Shiftlt|Shiftrt|BAnd), e1, e2, _) ->
	let te1pre, te1post, _, _, c1 = f_e loc rw e1 in
	let te2pre, te2post, _, _, c2 = f_e loc rw e2 in
	let prete, postte = tensor_tenv loc te1pre te2pre, tensor_tenv loc te1post te2post in
	  prete, postte, TSomething, TSomething, merge_cs [c1; c2; wf_te loc prete; wf_te loc postte]
    | BinOp(MinusPP, e1, e2, ciltyp) ->
	let te1pre, te1post, typ1pre, typ1post, c1 = f_e loc rw e1 in
	let te2pre, te2post, typ2pre, typ2post, c2 = f_e loc rw e2 in
	let prete, postte = tensor_tenv loc te1pre te2pre, tensor_tenv loc te1post te2post in
	  prete, postte, TSomething, TSomething, merge_cs [c1; c2; eq_typ loc typ1pre typ1post; eq_typ loc typ2pre typ2post; wf_te loc prete; wf_te loc postte]
    | BinOp((PlusPI|IndexPI|MinusPI), e1, e2, ciltyp) ->
	let te1pre, te1post, typ1pre, typ1post, c1 = f_e loc rw e1 in
	let te2pre, te2post, _, _, c2 = f_e loc rw e2 in
	  (* let typepre, typepost = rename_typ typ1pre, rename_typ typ1post in *)
	  (*
	    let c3 = merge_cs [(* empty_typ loc typepre; empty_typ loc typepost;  *)
	  (*eq_typ loc typ1pre (tensor_typ loc typ1post typepost)*)
	  (* eq_typ loc (tensor_typ loc typepre typ1pre) (tensor_typ loc typepost typ1post) *)] in
	  *)
	  (* [XXX] This is maybe not correct *)
	  (* tensor_tenv loc te1pre te2pre, tensor_tenv loc te1post te2post, typ1pre, typ1post, merge_cs [c1; c2]*)
	let prete, postte = tensor_tenv loc te1pre te2pre, tensor_tenv loc te1post te2post in
	  prete, postte, typ1pre, typ1post, merge_cs [c1; c2; wf_te loc prete; wf_te loc postte]
    | UnOp((Neg|LNot|BNot), e, _) ->
	f_e loc rw e
    | StartOf lv | Lval lv -> f_lv loc rw lv
    | SizeOfE e -> 
	let tepre, tepost, _, _, c = f_e loc rw e in
	  tepre, tepost, TSomething, TSomething, c
    | AddrOf lv ->
	(* [XXX] Do something if lv is taking offset of record. *)
	let ciltyp_lv = typeOf (Lval lv) in
	let tepre, tepost, tpre, tpost, c = f_lv loc rw lv in
	  if isIntegralType ciltyp_lv || isArithmeticType ciltyp_lv || isFunctionType ciltyp_lv then
	    tepre, tepost, (TRef(tpre, OFrac 0.0, None)), (TRef(tpost, OFrac 0.0, None)), c
	  else
	    tepre, tepost, (TRef(tpre, OFrac 0.0, Some (OFrac 1.0))), (TRef(tpost, OFrac 0.0, Some (new_ovar ()))), c
    | CastE(typ, e) -> 
	(*
	  let tepre, tepost, _, _, c = f_e loc rw e in
	  tepre, tepost, typ_of_ciltyp typ, typ_of_ciltyp typ, c
	*)
	let ciltyp_e = typeOf e in
	let tepre, tepost, tpre, tpost, cs1 = f_e loc rw e in
	  (* [XXX] deal with void* *)
	  if isZero e then
	    ((*fstwarn "unconstrained@\n";*)
	      tepre, tepost, typ_of_ciltyp typ, typ_of_ciltyp typ, cs1)
	  else if isVoidPtrType typ || (ciltyp_iscompatible ciltyp_e typ) then
	    tepre, tepost, tpre, tpost, cs1
	  else if isVoidPtrType ciltyp_e then
	    let cs2 = merge_cs [eq_typ loc tpre (typ_of_ciltyp typ); eq_typ loc tpost (typ_of_ciltyp typ)] in
	      tepre, tepost, tpre, tpost, union_cs cs1 cs2
	  else
	    (fstwarn "<typ>@[%a@]@\n<e>@[%a@]@\n<typeOf e>@[%a@]@\n<loc>@[%a@]@\n" print_ciltyp typ print_exp e print_ciltyp (typeOf e) print_loc loc;
	     failwith "f_e loc: Illegal cast")
    | _ -> 
	fstwarn "@[%a@]@\n" print_exp e;
	failwith "f_e: Not implmeneted yet"
    
let rec f file =
  let c1 = 
    List.fold_left
      (fun cs g -> union_cs cs (f_glob (* glob_te *) g))
      empty_cs
      file.globals
  in
  let loc = !currentLoc in
  let c2 = List.fold_left (fun cs (_, typ) -> CS.union cs (wf_typ loc typ)) empty_cs !globalTEnv in
  let collapse_subtype cs =
    let cs_r = ref CS.empty in
      CS.iter
	(fun c ->
	   match c with
	       CSubtype(TVar tv, typ, loc) ->
		 let cs = eq_typ loc (TVar tv) typ in
		   cs_r := CS.union cs !cs_r
	     | _ -> cs_r := CS.add c !cs_r)
	cs;
      !cs_r
  in
  let cs = merge_cs [c1; c2] in
  let cs = measure_and_print "collapse" (fun () -> normalize_cs (collapse_subtype cs)) () in
  let old_tmaxid = get_tmaxid () in
  let _ = printf "num constr: %d@\n%!" (CS.cardinal cs) in
  let cs = 
    measure_and_print 
      "expand" 
      (fun () -> expand_until_saturate (!currentLoc) old_tmaxid cs) () 
  in
  (* let _ = printf "num constr: %d@\n%!" (CS.cardinal cs) in *)
(*
    if !FreeSafeTyFlag.debug then
      dump_instr_hash ();
*)
(*
      (fstdebug "<global_ret>:@[%a@]@\n" print_tenv !globalTEnv;
       fstdebug "<global_cs>:@[%a@]@\n" print_cs cs);
*)
    !globalTEnv, cs

