
open Cil
open FreeSafeTyUtil
open FreeSafeTyGenConstr
open Format
open Graph

type ('varid, 'color, 'edge) node =
    { mutable color : 'color;
      mutable innodes : ('edge * 'varid) list;
      mutable outnodes : ('edge * 'varid) list }
type ('varid, 'color, 'edge) graph = ('varid, 'color, 'edge) node array
type edge = Eq | Le
type owner_color = Frac of float | Retracted of int | Nocolor | Notused

let rec print_owner_color fmt c = match c with
  | Frac f -> fprintf fmt "%g" f
  | Retracted n -> fprintf fmt "'n%d" n
  | Notused | Nocolor -> ()
let rec print_owner_edge fmt e = match e with
  | Eq -> fprintf fmt "====>"
  | Le -> fprintf fmt "---->"
let rec print_owner_graph fmt g =
  Array.iteri
    (fun i n ->
       match n.color with
	 | Retracted _ | Frac _ -> fprintf fmt "%d:@[%a@]@\n" i print_owner_color n.color
	 | Notused -> ()
	 | Nocolor ->
	     fprintf fmt "@[%a@]@\n" 
	       (print_list 
		  (fun fmt (e, i') -> fprintf fmt "%d@[%a@]%d" i print_owner_edge e i')
		  ~sep:"@\n")
	       n.outnodes)
    g

	
let make_empty_node color = { color = color; innodes = []; outnodes = [] }

module OVarIG = Imperative.Digraph.Concrete(
  struct
    type t = ovarid
    let compare ov1 ov2 = compare ov1.id ov2.id
    let hash ov = ov.id
    let equal ov1 ov2 = ov1.id = ov2.id
  end
)
module OVarSCC = Components.Make(OVarIG)
module TVarIG = Imperative.Digraph.Concrete(
  struct
    type t = tvarid
    let compare tv1 tv2 = compare (fst tv1.id) (fst tv2.id)
    let hash tv = fst tv.id
    let equal tv1 tv2 = (fst tv1.id) = (fst tv2.id)
  end
)
module TVarSCC = Components.Make(TVarIG)

let rec owner_constr_set_to_sccs cs =
  let module G = OVarIG in
  let g = G.create () in
    CS.iter
      (function CLeOwner(o1, o2, _) ->
	 let o1, o2 = find_owner o1, find_owner o2 in
	 begin match o1, o2 with
	   | OVar ov1, OVar ov2 ->
	       G.add_vertex g ov1;
	       G.add_vertex g ov2;
	       G.add_edge g ov1 ov2
	   | _ -> () end
	 | _ -> ())
      cs;
    OVarSCC.scc_list g
let rec tvar_constr_set_to_sccs cs =
  let module G = TVarIG in
  let g = G.create () in
    CS.iter
      (function CSubtype(typ1, typ2, _) ->
	 let typ1, typ2 = find_typ typ1, find_typ typ2 in
	   begin match typ1, typ2 with
	     | TVar tv1, TVar tv2 ->
		 assert(tv1.cont = None);
		 assert(tv2.cont = None);
		 G.add_vertex g tv1;
		 G.add_vertex g tv2;
		 G.add_edge g tv1 tv2;
	     | _ -> () end
	 | _ -> ())
      cs;
    TVarSCC.scc_list g

let rec retract_zero_one cs =
  let module G = OVarIG in
  let zero, one = new_ovar (), new_ovar () in
  let zero, one = match zero, one with OVar ov1, OVar ov2 -> ov1, ov2 | _ -> failwith "Cannot occur" in
  let g = G.create () in
  let rec put_edge o1 o2 =
    match o1, o2 with
      | OVar ov, OFrac 0.0 -> G.add_edge g ov zero 
      | OFrac 1.0, OVar ov -> G.add_edge g one ov
      | OVar ov1, OVar ov2 -> G.add_edge g ov1 ov2
      | OPlus os, OVar _ -> List.iter (fun o -> put_edge o o2) os
      | _ -> ()
  in
  let dirty = ref false in
    CS.iter
      (function CLeOwner(o1, o2, _) -> put_edge (find_owner o1) (find_owner o2)
	 | _ -> ())
      cs;
    if G.mem_vertex g zero then
      G.iter_pred
	(fun ov -> ignore(union_owner (!Cil.currentLoc) ov (OFrac 0.0)); dirty := true)
	g zero;
    if G.mem_vertex g one then
      G.iter_succ
	(fun ov -> ignore(union_owner (!Cil.currentLoc) ov (OFrac 1.0)); dirty := true)
	g one;
    cs, !dirty

 
let rec f (cs : constr_set) : constr_set =
  (* let cs = expand_once ~strong_expand:true (normalize_cs cs) in *)
  (* let _ = fstdebug "<*** BEFORE CS ***>@[%a@]@\n" print_cs cs in *)
  (* let _ = fstdebug "Doing preproc@\n%!" in *)
  let rec do_scc_owner cs =
    let _ = if !FreeSafeTyFlag.debug then fstdebug "<size of cs>@[%d@]@\n" (CS.cardinal cs) in
    let cs, dirty1 = if !FreeSafeTyFlag.noZeroOne then cs, false else retract_zero_one (normalize_cs cs) in
    let sccs = if !FreeSafeTyFlag.noSCC then [] else owner_constr_set_to_sccs (normalize_cs cs) in
    let dirty =
      List.fold_left
	(fun dirty scc ->
	   match scc with
	     | [] -> failwith "empty scc"
	     | [hd] -> dirty
	     | hd::tl ->
		 assert(List.for_all (fun ov -> ov.cont = None) scc);
		 List.iter 
		   (fun ov -> ignore(union_owner (!Cil.currentLoc) ov (OVar hd))) tl;
		 true)
	false
	sccs
    in
      if dirty1 || dirty then do_scc_owner (normalize_cs cs) else (normalize_cs cs)
  in
  let rec do_scc_typ cs =
    let _ = if !FreeSafeTyFlag.debug then fstdebug "<size of cs>@[%d@]@\n" (CS.cardinal cs) in
    (* let cs, dirty1 = if !FreeSafeTyFlag.noZeroOne then cs, false else retract_zero_one (normalize_cs cs) in *)
    let sccs = if !FreeSafeTyFlag.noSCC then [] else tvar_constr_set_to_sccs (normalize_cs cs) in
    let dirty =
      List.fold_left
	(fun dirty scc ->
	   match scc with
	     | [] -> failwith "empty scc"
	     | [hd] -> dirty
	     | hd::tl ->
		 assert(List.for_all (fun tv -> tv.cont = None) scc);
		 List.iter 
		   (fun tv -> ignore(union_typ (!Cil.currentLoc) tv (TVar hd))) tl;
		 true)
	false
	sccs
    in
      if dirty then do_scc_typ (normalize_cs cs) else (normalize_cs cs)
  in
  let cs = measure_and_print "scc" (fun () -> normalize_cs (do_scc_owner (do_scc_typ cs))) () in
  (* let cs = measure_and_print "scc" (fun () -> normalize_cs (do_scc_owner cs)) () in *)
    (*
      let cs = measure_and_print "collapse" (fun () -> normalize_cs (collapse_subtype cs)) () in
    (* [XXX] From here. If the following let... is removed, what makes dl-insert with "free(elm)" well-typed? *)
      let old_tmaxid = get_tmaxid () in
      let cs = measure_and_print "2nd" (fun () -> normalize_cs (expand_tvar_and_flatten (!currentLoc) old_tmaxid cs)) () in
      let cs = measure_and_print "scc" (fun () -> normalize_cs (do_scc_owner cs)) () in
    *)
    (* let cs = normalize_cs cs in *)
    (* let cs = helper cs in *)
    (* fstdebug "<*** After preproc ***>@\n"; *)
    (* dump_instr_hash (); *)
    cs
