
open FreeSafeTyFlag
open Cil
open Format

let print_loc fmt loc = Format.fprintf fmt "%s" (CilPretty.sprint ~width:80 (d_loc () loc))
(*
  Returns (lvst, lvsf)
  lvst is a list of lvals that must be null when e is evaluated to true (non-zero).
  lvsf is a list of lvals that must be null when e is evaluated to false (zero).
*)
let rec nullLvals e =
  match e with
    | Lval lv -> [], [lv]
    | CastE(_, e) -> nullLvals e
    | BinOp(Eq, e1, e2, _) ->
	if isZero e1 then
	  let lvs1, lvs2 = nullLvals e2 in
	    lvs2, lvs1
	else if isZero e2 then
	  let lvs1, lvs2 = nullLvals e1 in
	    lvs2, lvs1
	else
	  [], []
    | BinOp(Ne, e1, e2, _) ->
	if isZero e1 then
	  let lvs1, lvs2 = nullLvals e2 in
	    lvs1, lvs2
	else if isZero e2 then
	  let lvs1, lvs2 = nullLvals e1 in
	    lvs1, lvs2
	else
	  [], []
    | UnOp(LNot, e, _) ->
	let lvst, lvsf = nullLvals e in
	  lvsf, lvst
    | Const _ | SizeOf _ | SizeOfE _ | SizeOfStr _
    | AlignOf _ | AlignOfE _ | AddrOf _ | StartOf _ 
    | UnOp _ | BinOp _ -> [], []

(*
  Hash table which masp strings into something
*)
module StringMap = Hashtbl.Make(
  struct 
    type t = (* string *) string
    let equal = (=)
    let hash = Hashtbl.hash
  end
)

module StringSet = Set.Make(
  struct
    type t = string
    let compare = compare
  end
)

let rec print_list printer ~sep fmt l =
  match l with
    | [] -> ()
    | [hd] -> fprintf fmt "%a" printer hd
    | hd::tl -> 
	fprintf fmt "%a" printer hd;
	fprintf fmt sep;
	fprintf fmt "%a" (print_list printer ~sep:sep) tl

let rec sprint_list printer ?(separator=" ") () l =
  match l with
    | [] -> ""
    | [hd] -> Format.sprintf "%a" printer hd
    | hd::tl -> Format.sprintf "%a%s%a" printer hd separator (sprint_list printer ~separator:separator) tl

let print_opt printer fmt opt =
  match opt with
  | None -> fprintf fmt "NONE"
  | Some x -> fprintf fmt "%a" printer x

let compare_opt decide_eq s1opt s2opt =
  match s1opt, s2opt with
  | None, _ | _, None -> false
  | Some(s1), Some(s2) -> decide_eq s1 s2


let print_bytes_cnt = ref (-1)
let print_bytes fmt s =
  incr print_bytes_cnt;
  fprintf fmt "%s, %d: %f\n" s (!print_bytes_cnt) (Gc.allocated_bytes ())

let gc_msg b =
  if b then begin
    Gc.set { (Gc.get ()) with Gc.verbose = 0x3ff };
    flush stdout;
    flush stderr
  end else begin
    Gc.set { (Gc.get ()) with Gc.verbose = 0x000 };
    flush stdout;
    flush stderr
  end

let comb list1 list2 =
  List.fold_right
    (fun elm1 res ->
      (List.map (fun elm2 -> (elm1, elm2)) list2) @ res)
    list1
    []

let rec fold_left_ofs (h : 'a) f_field f_index ofs =
  match ofs with
    | NoOffset -> h
    | Field(f, ofs') ->
	let h' = f_field h f in
	  fold_left_ofs h' f_field f_index ofs'
    | Index(e, ofs') ->
	let h' = f_index h e in
	  fold_left_ofs h' f_field f_index ofs'

let cnt = ref (-1)
let new_int () =
  (incr cnt; !cnt)

let partition_list compare f_common l1 l2 =
  let l1', l2' = (List.sort compare l1), (List.sort compare l2) in
  let rec help only_in_l1 only_in_l2 common l1 l2 =
    match l1, l2 with
      | [], [] -> (only_in_l1, only_in_l2, common)
      | [], l -> (only_in_l1, ((List.rev only_in_l2) @ l), common)
      | l, [] -> (((List.rev only_in_l1) @ l), only_in_l2, common)
      | (hd1::tl1), (hd2::tl2) ->
	  let cmp = compare hd1 hd2 in
	    if cmp = 0 then
	      help only_in_l1 only_in_l2 ((f_common hd1 hd2)::common) tl1 tl2
	    else if cmp < 0 then
	      help (hd1::only_in_l1) only_in_l2 common tl1 l2
	    else
	      help only_in_l1 (hd2::only_in_l2) common l1 tl2
  in
    help [] [] [] l1' l2'

let diff_list compare l1 l2 =
  let l, _, _ = partition_list compare (fun hd1 hd2 -> ()) l1 l2 in
    l

let common_list compare l1 l2 =
  let _, _, common = partition_list compare (fun hd1 hd2 -> hd1) l1 l2 in
    common

let rec safe_map2 f l1 l2 =
  match l1, l2 with
    | [], _ | _, [] -> []
    | hd1::tl1, hd2::tl2 ->
	(f hd1 hd2)::(safe_map2 f tl1 tl2)

let rec safe_iter2  f l1 l2 =
  match l1, l2 with
    | [], _ | _, [] -> ()
    | hd1::tl1, hd2::tl2 ->
	(f hd1 hd2); (safe_iter2 f tl1 tl2)

let get_fstyfile_name s =
  (Filename.chop_extension s) ^ ".fst"
let get_globalfile_name s =
  (Filename.chop_extension s) ^ ".gfst"

module ImperativeSet(E : Hashtbl.HashedType) :
  (sig
     type t
     type elm = E.t
     val create : int -> t
     val add : elm -> t -> unit
     val iter : (elm -> unit) -> t -> unit
     val pop : t -> elm
     val cardinal : t -> int
     val filter : (elm -> bool) -> t -> t
       (**
	  filter f h returns elements in h which satisfy the predicate f
	  and destructively subtract returned elements from h.
       *)
     val filter_copy : (elm -> bool) -> t -> t
     val partition_copy : (elm -> bool) -> t -> (t * t)
     val clear_and_set : t -> t -> unit
       (**
	  clear_and_set h1 h2 clears h1 and destructively copy elements in h2 to h1.
       *)
     val union_update : t -> t -> unit
       (**
	  union_update h1 h2 destructively makes h1 the union of h1 and h2.
       *)
   end)
  =
  (struct
     module H = Hashtbl.Make(E) 
     exception Local of H.key
     type t = unit H.t
     type elm = E.t

     let add k h = 
       (* Format.printf "add <1>: %d\n" (H.length h); *)
       let _ =
	 if H.mem h k then
	   ()
	 else
	   H.add h k ()
       in
	 (* Format.printf "add <2>: %d\n" (H.length h) *)
	 ()

     let create = H.create

     let pop h =
       (* Format.printf "pop <1>: %d\n" (H.length h); *)
       if H.length h = 0 then 
	 raise Not_found
       else
	 try 
	   H.iter (fun key _ -> raise (Local key)) h;
	   failwith "impossible"
	 with Local x -> 
	   H.remove h x;
	   (* Format.printf "pop <2>: %d\n" (H.length h); *)
	   x

     let iter f h =
       H.iter (fun key _ -> f key) h

     let cardinal h = H.length h

     let partition_copy f h =
       let true_h = H.create 128 in
       let false_h = H.create 128 in
	 H.iter (fun elm _ -> if f elm then H.add true_h elm () else H.add false_h elm ()) h;
	 (true_h, false_h)

     let filter_copy f h =
       fst (partition_copy f h)

     let clear_and_set h1 h2 =
       H.clear h1;
       H.iter (fun elm () -> H.add h1 elm ()) h2

     let filter f h =
       let true_h, false_h = partition_copy f h in
	 clear_and_set h false_h;
	 true_h

     let union_update h1 h2 =
       H.iter (fun elm _ -> H.add h1 elm ()) h2
   end)

let measure_time f c =
  let prev = Unix.gettimeofday () in
  let x = f c in
  let post = Unix.gettimeofday () in
    (x, post -. prev)

let measure_memory f c =
  let prev = Gc.allocated_bytes () in
  let x = f c in
  let post = Gc.allocated_bytes () in
    (x, post -. prev)

let measure_and_print prefix f c =
  let (c, time), memory =
    measure_memory
      (fun c ->
	 let c, time = measure_time f c in
	   c, time)
      c
  in
  let _ = 
    (* if !measureTime then *)
    Format.printf "%s: " prefix;
    Format.printf "%g sec, %g bytes@\n%!" time memory;
    print_flush ()
  in
    c

let rec seqapp l c =
  let rec helper l res =
    match l with
	[] -> res
      | hd::tl ->
	  helper tl (hd res)
  in
    helper l c

module LocSet = Set.Make(
  struct
    type t = location
    let compare = compare
  end
)
let print_locset fmt locs = fprintf fmt "@[%a@]" (print_list print_loc ~sep:";") (LocSet.elements locs)
let empty_locs = LocSet.empty
type ('cont, 'b) tvarinfo =
    { id : 'b;
      mutable cont : 'cont option;
      mutable locs : LocSet.t }
(* let maxid = ref (-1) *)
let new_tvar id (* idtypetype *) =
  (* incr maxid;  *)
  (* if !maxid = 81875 || !maxid = 81881 then (printf "Utils.new_tvar:
     %d is generated@\n" !maxid; flush stdout);
  *)
  { id = id (* (fileId, idtypetype, !maxid) *); cont = None; locs = empty_locs }
let print_tvar cont_printer print_id prefix fmt tv =
  match tv.cont with
    | None -> fprintf fmt "'%s[%a]" prefix print_id tv.id
    | Some cont ->
	(* fprintf fmt "%a(='%s[%a])" cont_printer cont prefix print_id tv.id *)
	fprintf fmt "%a" cont_printer cont
let compare_tvarinfo x1 x2 = compare x1.id x2.id
let sort_tvars vs = List.sort compare_tvarinfo vs
let mem_novisits novisits tvi =
  List.mem tvi.id novisits
  (* List.exists ((=) tvi.id) novisits *)
let rec find (ret_tvaropt : 'a -> ('a, 'b) tvarinfo option) lift v =
  let rec find nofind v =
    match v.cont with
      | None -> (lift v, v.locs)
      | Some t -> 
	  begin
	    match ret_tvaropt t with
	      | None -> (t, v.locs)
	      | Some tv ->
		  (* if List.mem tv.id nofind then *)
		  if List.exists ((==) tv) nofind then
		    t, v.locs
		  else
		    (* let newcont, locs = find (tv.id::nofind) tv in *)
		    let newcont, locs = find (tv::nofind) tv in
		      v.cont <- Some(newcont);
		      v.locs <- LocSet.union v.locs locs;
		      newcont, v.locs
	  end
  in
    (* fst (find (* [v.id] *) [v] v) *)
    find [v] v

let rec merge locs occurcheck empty (* f *) tv t =
  if occurcheck tv t then
    (empty, false)
  else
    match tv.cont with
      | Some t' -> failwith "find is not appropreately performed" (* (f t t', true)*)
      | None ->
	  tv.locs <- LocSet.union locs tv.locs;
	  tv.cont <- Some(t); 
	  (empty, true)

let is_struct_ti ti = 
  match unrollType ti.ttype with TComp(ci, _) -> ci.cstruct
    | _ -> false
let is_union_ti ti = 
  match unrollType ti.ttype with TComp(ci, _) -> not ci.cstruct
    | _ -> false

let rec compare_delayed_list l =
  match l with
    | [] ->  0
    | hd::tl ->
	let c = hd () in
	  if c = 0 then compare_delayed_list tl
	  else c

let rec compare_list cmp sort l1 l2 =
  let l1 = sort l1 in
  let l2 = sort l2 in
  let rec help l1 l2 =
    match l1, l2 with
      | [], [] -> 0
      | [], _ -> -1
      | _, [] -> 1
      | hd1::tl1, hd2::tl2 ->
	  compare_delayed_list
	    [(fun () -> cmp hd1 hd2);
	     (fun () -> help tl1 tl2)]
  in
    help l1 l2
(*
    compare_delayed_list
      (List.map2 (fun elm1 elm2 () ->  cmp elm1 elm2) l1 l2)
*)
    

let hash_list l = Hashtbl.hash l
let hash_fold f l = hash_list (List.map f l)
let hash_string s = Hashtbl.hash s

let rec last_list l =
  match l with
    | [] -> failwith "last_list"
    | [elm] -> elm
    | hd::tl -> last_list tl

module CilVarSet = Set.Make(
  struct
    type t = Cil.varinfo
    let compare vi1 vi2 = compare vi1.vid vi2.vid
  end
)
type cilvarset = CilVarSet.t
let print_cilvarset fmt vs =
  CilVarSet.iter
    (fun vi ->
       Format.printf "%s," vi.vname)
    vs

let isFunctionType ciltyp =
  let ciltyp' = unrollType ciltyp in
    match ciltyp' with
      | TFun _ | TPtr(TFun _, _) -> true
      | _ -> false

let uniq_append compare l1 l2 =
  let only1, only2, common =
    partition_list 
      compare
      (fun v _ -> v)
      l1 l2
  in
    only1 @ only2 @ common
(*
  let uniq_append fgvs1 fgvs2 =
    let only1, only2, common =
      partition_list 
	(fun fgv1 fgv2 -> compare (id_of_freegvvarinfo fgv1) (id_of_freegvvarinfo fgv2))
	(fun fgv _ -> fgv)
	fgvs1 fgvs2
    in
      only1 @ only2 @ common
  in
*)

let fstwarn format = printf ("<warning>" ^^ format)
let fstdebug format = printf ("<debug>" ^^ format)

open CilPretty
let pr = defaultCilPrinter
let print_varinfo fmt vi = Format.fprintf fmt "%s" (sprint ~width:80 (pr#pVar vi))
let print_ciltyp fmt typ = Format.fprintf fmt "%s" (sprint ~width:80 (d_type () typ))
let print_stmt fmt s = Format.fprintf fmt "%s" (sprint ~width:80 (d_stmt () s))
(* let print_block fmt b = Format.fprintf fmt "<block>%s</block>" (sprint ~width:80 (d_block () b)) *)
let print_block fmt b = Format.fprintf fmt "%s" (sprint ~width:80 (d_block () b))
let print_fundec fmt fd = Format.fprintf fmt "%s" (sprint ~width:80 (d_block () fd.sbody))
let print_offset fmt ofs = Format.fprintf fmt "%s" (sprint ~width:80 (d_offset CilPretty.nil () ofs))
let print_exp fmt e = Format.fprintf fmt "%s" (sprint ~width:80 (d_exp () e))
let print_lval fmt lv = Format.fprintf fmt "%s" (sprint ~width:80 (d_lval () lv))
let print_instr fmt i = Format.fprintf fmt "%s" (sprint ~width:80 (d_instr () i))


let rec stripCastChain e =
  match e with
    | CastE(_, e) -> stripCastChain e
    | Lval lv -> lv
    | _ ->
	fstwarn "@[%a@]@\n" print_exp e;
	failwith "stripCastChain: applied to non-lval expression."

let rec add_ofs lv ofs =
  let rec add_ofs_to_ofs' ofs' ofs =
    match ofs' with
      | NoOffset -> ofs
      | Field(f, ofs'') -> Field(f, add_ofs_to_ofs' ofs'' ofs)
      | Index(e, ofs'') -> Index(e, add_ofs_to_ofs' ofs'' ofs)
  in
  match lv with
    | h, NoOffset -> h, ofs
    | h, Field(f, ofs') ->
	h, Field(f, add_ofs_to_ofs' ofs' ofs)
    | h, Index(e, ofs') ->
	h, Index(e, add_ofs_to_ofs' ofs' ofs)

let rec ciltyp_iscompatible ciltyp1 ciltyp2 =
  match unrollType ciltyp1, unrollType ciltyp2 with
    | (TVoid _ | TInt _ | TFloat _ | TEnum _), (TVoid _ | TInt _ | TFloat _ | TEnum _) -> true
    | (TPtr(typ1, _) | TArray(typ1, _, _)) , (TPtr(typ2, _) | TArray(typ2, _, _)) ->
	ciltyp_iscompatible typ1 typ2
    | TComp(ci1, _), TComp(ci2, _) -> ci1.cname = ci2.cname
    | _ -> false

let rec safe_fold_left f t ts1 ts2 =
  match ts1, ts2 with
    | [], _ | _, [] -> t
    | hd1::tl1, hd2::tl2 ->
	safe_fold_left f (f t hd1 hd2) tl1 tl2

let split_random l =
  List.fold_left
    (fun (l1, l2) x ->
       if Random.bool () then
	 (x::l1, l2)
       else
	 (l1, x::l2))
    ([], [])
    l
