

open Cil
open CilPretty

module E = Errormsg

(***
    NormalizeLvalVisitor:
    Normalize lvalues by making dereferencing explicit.
    More precisely, each lvalue of type T is converted into
    an lvalue of type T*. At the same time, a variable of
    type T is defined on the stack.

    Example:
    x = y + z 
    ===>
    int x';
    int y';
    int z';
    int* x = &x';
    int* y = &y';
    int* z = &z';
    ...
    *x = *y + *z

    Result of the translation is described by the following
    where "var" is the type of lval of the form (Var x, NoOffset)
    and "memlval" is the type of lval of the form (Mem x, NoOffset).

    type exp =
    Const      of constant
    | Lval       of lval
    | SizeOf     of typ                   
    | SizeOfE    of lval
    | SizeOfStr  of string
    | AlignOf    of typ                   
    | AlignOfE   of lval
    | UnOp       of unop * lval * typ
    | BinOp      of binop * lval * lval * typ
    | CastE      of typ * lval
(* | AddrOf     of lval *)
(* | StartOf    of lval *)
    and lval = lhost * offset
    and lhost =
(* | Var of varinfo *)
    | Mem of var
    and offset =
    | NoOffset
    | Field of fieldinfo * offset (always NoOffset)
    | Index of lval * offset (always NoOffset)
    and stmtkind =
    | Switch of lval * block * (stmt list) * loc
    ...
    and instr =
(* x = &y: 
    occurs only the begining of 
    each function *)
    | SetAddrOf of var * var * loc
    | SetStartOf of var * var * loc
(* *x = &x->f... *)
    | AssignFieldAddrOf of mamlval * lval * loc
    | Set of lval * exp * loc
    | CallVar of lval option (always None) * var * lval list * loc
    | CallMem of lval option (always None) * lval * lval list * loc
    ...
***)

module VarinfoHash = Hashtbl.Make(
  struct
    type t = varinfo
    let equal = (==)
    let hash = Hashtbl.hash
  end)

module VarinfoSet = Set.Make(
  struct
    type t = varinfo
    let compare vi1 vi2 = compare vi1.vid vi2.vid
  end
)

(* let makeBodyName s = "__vericon_body_" ^ s *)
let makePtrName s = "__vericon_ptr_" ^ s

class type ['a] tobeProcessedVisitorType = object
  inherit nopCilVisitor
  method getResult : unit -> 'a
end

class tobeProcessedVisitor : [VarinfoSet.t] tobeProcessedVisitorType = object(self)
  inherit nopCilVisitor
    
  val set = ref VarinfoSet.empty

  method vinst i =
    match i with
      | Set((Var x, NoOffset), e, loc) when x.vglob = false ->
	  ignore(visitCilExpr (self :> cilVisitor) e);
	  set := VarinfoSet.add x !set;
	  SkipChildren
      | _ -> DoChildren

  method vexpr e =
    match e with
      | AddrOf(Var x, NoOffset) when x.vglob = false ->
	  set := VarinfoSet.add x !set;
	  SkipChildren
      | _ -> DoChildren

  method getResult () = !set
end

class normalizeLvalVisitor : cilVisitor = object(self)
  inherit nopCilVisitor

  val varinfoGlobalHash = VarinfoHash.create 128
  val varinfoLocalHash = VarinfoHash.create 128
    
  method private queryVarinfoHash vi =
    if vi.vglob then
      VarinfoHash.find varinfoGlobalHash vi
    else
      VarinfoHash.find varinfoLocalHash vi      

  val currentFundec = ref dummyFunDec
  val initInstrs = ref []

  (* If vi.vtype is a function type, then this is a prototype. *)
  (* We cannot initialize nor take address of register variables. *)
  method private shouldnotBeProcessed vi =
    (* [XXX] What about functional pointers that are not globals?*)
    (not vi.vaddrof && not vi.vglob)
    || isFunctionType vi.vtype 
    || vi.vstorage = Register 

  method private rename_section vi =
    vi.vattr <-
      List.map 
      (function 
	 | (Attr("section", [AStr s])) ->
	     Attr("section", [AStr ("_vericon_" ^ s)])
	 | att -> att)
      vi.vattr

  method private processGlobalVar vi loc f =
    if self#shouldnotBeProcessed vi then
      SkipChildren
    else
      let newvi = copyVarinfo vi (makePtrName vi.vname) in
	newvi.vtype <- TPtr(vi.vtype, []);
	self#rename_section newvi;
	VarinfoHash.add varinfoGlobalHash vi newvi;
	(*
	  let newvi' = copyVarinfo newvi (makePtrName newvi.vname) in
	  newvi'.vtype <- TPtr(newvi.vtype, []);
	  self#rename_section newvi';
	  VarinfoHash.add varinfoGlobalHash newvi newvi';
	*)
	ChangeTo (f vi newvi loc)

  method vexpr e =
    let processVar x =
      if self#shouldnotBeProcessed x then
	SkipChildren
      else
	let x' = 
	  self#queryVarinfoHash x
	in
	  (* ChangeTo(Lval(Mem(Lval(var x')), NoOffset)) *)
	  ChangeTo(Lval(Var x', NoOffset))
    in
      match e with
	| AddrOf(Var x, NoOffset) -> processVar x
	| StartOf(Var x, NoOffset) -> (* processVar x *)
	    (* ChangeTo(StartOf(Mem(Lval(Var (self#queryVarinfoHash x), NoOffset)), NoOffset)) *)
	    SkipChildren
	| StartOf(lv, ofs) -> DoChildren
	| AddrOf(_, ofs) -> DoChildren
	    (* | AlignOf _ -> SkipChildren *)
	| _ -> DoChildren

  method vfunc fd =
    currentFundec := fd;
    assert(VarinfoHash.length varinfoLocalHash = 0);
    let tobeProcessedVisitor = new tobeProcessedVisitor in
    let _ = visitCilBlock (tobeProcessedVisitor :> cilVisitor) fd.sbody in
    let vi_set = tobeProcessedVisitor#getResult () in
    let setinstrs = ref [] in
      List.iter
	(fun vi ->
	   if not (self#shouldnotBeProcessed vi) then
	     let newvi = makeLocalVar fd (makePtrName vi.vname) (TPtr(vi.vtype, [])) in
	       (* let newvi' = makeLocalVar fd (makePtrName newvi.vname) (TPtr(newvi.vtype, [])) in*)
	       VarinfoHash.add varinfoLocalHash vi newvi;
	       (* VarinfoHash.add varinfoLocalHash newvi newvi';*)
	       setinstrs := 
		 (Set(var newvi, mkAddrOf(var vi), !currentLoc))::!setinstrs)
	(* ((Set(var newvi', mkAddrOf(var newvi), !currentLoc)):: *)
	(VarinfoSet.elements vi_set);
      let b = visitCilBlock (self :> cilVisitor) fd.sbody in
	(* We postpone to add setstmts to fd.sbody here to avoid &x ==> x conversion *)
	(match !setinstrs with [] -> () | _ -> fd.sbody.bstmts <- ((mkStmt(Instr(List.rev (!setinstrs)))) :: b.bstmts));
	VarinfoHash.clear varinfoLocalHash;
	SkipChildren
	  (*
	  (* Need to call setFormals with fd.sformals to adjust fd.vtype. *)
	    setFormals fd fd.sformals;
	  *)
	  
  method vinst = function
    | Asm _ -> SkipChildren
	(*
	  | Set((Mem(Lval(Var x, NoOffset)), NoOffset) as lv,
	  e,
	  loc) ->
	  let e' = visitCilExpr (self :> cilVisitor) e in
	  ChangeTo([Set(lv, e', loc)])
	  | Set((Var x, NoOffset) as lv,
	  e,
	  loc) ->
	  let lv' = visitCilLval (self :> cilVisitor) lv in
	  let e' = visitCilExpr (self :> cilVisitor) e in
	  ChangeTo([Set(lv', e', loc)])
	*)
    | _ -> DoChildren

  method vlval (h, ofs) =
    let _ =
      if !currentLoc.line = 166 then
	()
    in
    let h' = 
      match h with
	  Var x -> 
	    if self#shouldnotBeProcessed x then
	      Var x
	    else begin
	      try
		Mem(Lval(Var(self#queryVarinfoHash x), NoOffset))
	      with
		| Not_found -> Var x
	    end
	      (* | Mem(Lval(Var x, NoOffset)) -> h *)
	| Mem(e) -> Mem(visitCilExpr (self :> cilVisitor) e)
    in
    let ofs' =
      match ofs with
	| Index(e, ofs') -> Index(visitCilExpr (self :> cilVisitor) e, visitCilOffset (self :> cilVisitor) ofs')
	| Field(f, ofs') -> Field(f, visitCilOffset (self :> cilVisitor) ofs')
	| NoOffset -> NoOffset
	    (*
	      Index(Lval((Var x, NoOffset) as lv), NoOffset) -> Index(Lval(visitCilLval (self :> cilVisitor) lv), NoOffset)
	      | Field(f, ofs') when f.fbitfield <> None -> (* ofs *)
	      | _ -> ofs
	    *)
    in
      ChangeTo(h', ofs')

  method vglob = function
    | GVarDecl(vi, loc) ->
	self#processGlobalVar vi loc
	  (fun vi newvi loc -> [GVarDecl(vi, loc); GVarDecl(newvi, loc)])
    | GVar(vi, init, loc) ->
	self#processGlobalVar vi loc
	  (fun vi newvi loc ->
	     [GVar(vi, init, loc); GVar(newvi, {init = Some (SingleInit(mkAddrOf(var vi)))}, loc);
	      (* GVar(newvi', {init = Some (SingleInit(mkAddrOf(var newvi)))}, loc) *)])
    | GFun(fd, loc) -> DoChildren
    | _ -> SkipChildren
end
