
open FreeSafeTyFlag
open Cil
open CilPretty

module E = Errormsg

module VarinfoSet = Set.Make(
  struct
    type t = varinfo
    let compare = compare
  end
)

(***
    TuneFunctionType:
    - Every function is converted to one that returns nothing.
      i.e. a function of the type (t1, t2, ..., tn) -> t is converted to on
           of the type (t*, t1, t2, ..., tn) -> void
***)

let ret_varname = "__fst_ret"

let rec rettyp = function
  | TFun(typ, _, _, _) -> typ
  (* | TPtr(typ, _) -> rettyp typ *)
  | _ -> voidType

(*
  - return e ===> *r = e; return
  - f(e1, e2, ..., en) ===> f(&r, e1, e2, ..., en) when f is not a function that returns void
    (r is a local variable for receiving the result)
  - tres f(t1, ..., tn) ===> tres f(tres*, t1, ..., tn) when tres != void
  - Change names of builtin functions
*)
let is_builtin s = Str.string_match (Str.regexp "__builtin") s 0
let rename_builtin vi = if is_builtin vi.vname then vi.vname <- "__fst" ^ vi.vname

class tuneCallInstrVisitor : cilVisitor = object(self)
  inherit nopCilVisitor

  val currentFundec = ref dummyFunDec
  val currentRetVarinfo = ref None

  (*
    method private delete_section vi =
    vi.vattr <-
    List.map 
    (function 
    | (Attr("section", [AStr s])) ->
    Attr("section", [AStr ("_vericon_" ^ s)])
    | att -> att)
    vi.vattr
  *)

  method vvdec vi = 
    (* self#delete_section vi; *)
    rename_builtin vi; 
    DoChildren
  method vvrbl vi = rename_builtin vi; DoChildren

  method vinst = function
(*
    | Call(_, Lval(Var x, _), _, _) when x.vname = name_strip_and_return ->
	DoChildren
*)
    | Call(lvopt, e, es, loc) ->
	let e = visitCilExpr (self :> cilVisitor) e in
	let es = List.map (visitCilExpr (self :> cilVisitor)) es in
	let funtyp = unrollType (typeOf e) in
	let rtyp = rettyp funtyp in
	  if isVoidType rtyp then
	    DoChildren
	  else
	    (* The function returns something *)
	    begin
	      match lvopt with 
		  (* The return value is ignored.
		     Create a dummy variable for the return value. *)
		| None -> 
		    let lvretmem = var (makeTempVar !currentFundec ~name:"__retmem" rtyp) in
		      (* let lvret = var (makeTempVar !currentFundec (TPtr(rtyp, []))) in *)
		    let addrofexp = mkAddrOf lvretmem in
		      ChangeTo([Call(None, e, addrofexp::es, loc)])
		| Some lv -> 
		    let lv = visitCilLval (self :> cilVisitor) lv in
		      ChangeTo([Call(None, e, (mkAddrOf lv)::es, loc)])
	    end
    | _ -> DoChildren

  method private get_current_retvinfo () =
    match !currentRetVarinfo with
	None -> E.s "function with return type void returns something"
      | Some vi -> vi

  method vtype t =
    match unrollType t with
      | (TFun _) as typ ->
	  let _ =
	    if !currentFundec.svar.vname = "hash_function" then
	      ()
	  in
	    ChangeDoChildrenPost(
	      typ,
	      (function 
		 | (TFun(tret, ts, b, attr)) as origtyp ->
		     if isVoidType tret then
		       origtyp
		     else
		       let convtyp = TFun(tret, Some((ret_varname, TPtr(tret, []), [])::(argsToList ts)), b, attr) in
			 begin
			   match argsToList ts with
			     | [] -> convtyp
			     | (s, _, _)::tl when s = ret_varname -> origtyp
			     | _ -> convtyp
			 end
		 | _ -> E.s (bug "vtype: incompatible type: %a")))
      | _ -> DoChildren

  method vstmt s =
    match s.skind with
	Return(Some(retexp), loc) ->
	  let retexp = visitCilExpr (self :> cilVisitor) retexp in
	    self#queueInstr [Set(mkMem ~addr:(Lval(var (self#get_current_retvinfo ()))) ~off:NoOffset, retexp, loc)];
	    (* ChangeTo({ s with skind = Return(None, loc)}) *)
	    s.skind <- Return(None, loc);
	    SkipChildren
      | _ -> DoChildren

  method vfunc fd =
    currentFundec := fd;
    let rtyp = rettyp fd.svar.vtype in
      if isVoidType rtyp then
	begin
	  currentRetVarinfo := None;
	  DoChildren
	end
      else
	begin
	  let retvinfo = makeFormalVar fd ~where:"^" ret_varname (TPtr(rtyp, [])) in
	    currentRetVarinfo := Some(retvinfo);
	    DoChildren
	end

end

class tuneFunctionTypeVisitor : cilVisitor = object(self)
  inherit nopCilVisitor

  method vtype = function
    | (TFun(rtyp, _, _, _)) as typ ->
	if isVoidType rtyp then
	  DoChildren
	else
	  ChangeDoChildrenPost(
	    typ,
	    (fun typ ->
	       match typ with
		 | TFun(rtyp, argtyps, b, attrs) ->
		     TFun(voidType, argtyps, b, attrs)
		 | _ -> failwith "tuneFunctionTypeVisitor: Unexpected type."))
    | _ -> DoChildren
end
